# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import pytest

from hypothesis.errors import InvalidArgument
from hypothesis.extra.array_api import NominalVersion, make_strategies_namespace

from tests.array_api.common import MIN_VER_FOR_COMPLEX
from tests.common.debug import check_can_generate_examples


def e(name, *, _min_version: NominalVersion | None = None, **kwargs):
    kw = ", ".join(f"{k}={v!r}" for k, v in kwargs.items())
    id_ = f"{name}({kw})"
    if _min_version is None:
        marks = ()
    else:
        marks = pytest.mark.xp_min_version(_min_version)
    return pytest.param(name, kwargs, id=id_, marks=marks)


@pytest.mark.parametrize(
    ("strat_name", "kwargs"),
    [
        e("arrays", dtype=1, shape=5),
        e("arrays", dtype=None, shape=5),
        e("arrays", dtype="int8", shape=(0.5,)),
        e("arrays", dtype="int8", shape=1, fill=3),
        e("arrays", dtype="int8", shape=1, elements="not a strategy"),
        e("arrays", dtype="int8", shape="not a shape or strategy"),
        e("array_shapes", min_side=2, max_side=1),
        e("array_shapes", min_dims=3, max_dims=2),
        e("array_shapes", min_dims=-1),
        e("array_shapes", min_side=-1),
        e("array_shapes", min_side="not an int"),
        e("array_shapes", max_side="not an int"),
        e("array_shapes", min_dims="not an int"),
        e("array_shapes", max_dims="not an int"),
        e("from_dtype", dtype=1),
        e("from_dtype", dtype=None),
        e("from_dtype", dtype="int8", min_value="not an int"),
        e("from_dtype", dtype="int8", max_value="not an int"),
        e("from_dtype", dtype="float32", min_value="not a float"),
        e("from_dtype", dtype="float32", max_value="not a float"),
        e("from_dtype", dtype="int8", min_value=10, max_value=5),
        e("from_dtype", dtype="float32", min_value=10, max_value=5),
        e("from_dtype", dtype="int8", min_value=-999),
        e("from_dtype", dtype="int8", max_value=-999),
        e("from_dtype", dtype="int8", min_value=999),
        e("from_dtype", dtype="int8", max_value=999),
        e("from_dtype", dtype="uint8", min_value=-999),
        e("from_dtype", dtype="uint8", max_value=-999),
        e("from_dtype", dtype="uint8", min_value=999),
        e("from_dtype", dtype="uint8", max_value=999),
        e("from_dtype", dtype="float32", min_value=-4e38),
        e("from_dtype", dtype="float32", max_value=-4e38),
        e("from_dtype", dtype="float32", min_value=4e38),
        e("from_dtype", dtype="float32", max_value=4e38),
        e("integer_dtypes", sizes=()),
        e("integer_dtypes", sizes=(3,)),
        e("unsigned_integer_dtypes", sizes=()),
        e("unsigned_integer_dtypes", sizes=(3,)),
        e("floating_dtypes", sizes=()),
        e("floating_dtypes", sizes=(3,)),
        e("complex_dtypes", _min_version=MIN_VER_FOR_COMPLEX, sizes=()),
        e("complex_dtypes", _min_version=MIN_VER_FOR_COMPLEX, sizes=(3,)),
        e("valid_tuple_axes", ndim=-1),
        e("valid_tuple_axes", ndim=2, min_size=-1),
        e("valid_tuple_axes", ndim=2, min_size=3, max_size=10),
        e("valid_tuple_axes", ndim=2, min_size=2, max_size=1),
        e("valid_tuple_axes", ndim=2.0, min_size=2, max_size=1),
        e("valid_tuple_axes", ndim=2, min_size=1.0, max_size=2),
        e("valid_tuple_axes", ndim=2, min_size=1, max_size=2.0),
        e("valid_tuple_axes", ndim=2, min_size=1, max_size=3),
        e("broadcastable_shapes", shape="a"),
        e("broadcastable_shapes", shape=(2, 2), min_side="a"),
        e("broadcastable_shapes", shape=(2, 2), min_dims="a"),
        e("broadcastable_shapes", shape=(2, 2), max_side="a"),
        e("broadcastable_shapes", shape=(2, 2), max_dims="a"),
        e("broadcastable_shapes", shape=(2, 2), min_side=-1),
        e("broadcastable_shapes", shape=(2, 2), min_dims=-1),
        e("broadcastable_shapes", shape=(2, 2), min_side=1, max_side=0),
        e("broadcastable_shapes", shape=(2, 2), min_dims=1, max_dims=0),
        e(
            "broadcastable_shapes",  # max_side too small
            shape=(5, 1),
            min_dims=2,
            max_dims=4,
            min_side=2,
            max_side=3,
        ),
        e(
            "broadcastable_shapes",  # min_side too large
            shape=(0, 1),
            min_dims=2,
            max_dims=4,
            min_side=2,
            max_side=3,
        ),
        e(
            "broadcastable_shapes",  # default max_dims unsatisfiable
            shape=(5, 3, 2, 1),
            min_dims=3,
            max_dims=None,
            min_side=2,
            max_side=3,
        ),
        e(
            "broadcastable_shapes",  # default max_dims unsatisfiable
            shape=(0, 3, 2, 1),
            min_dims=3,
            max_dims=None,
            min_side=2,
            max_side=3,
        ),
        e("mutually_broadcastable_shapes", num_shapes=0),
        e("mutually_broadcastable_shapes", num_shapes="a"),
        e("mutually_broadcastable_shapes", num_shapes=2, base_shape="a"),
        e(
            "mutually_broadcastable_shapes",  # min_side is invalid type
            num_shapes=2,
            min_side="a",
        ),
        e(
            "mutually_broadcastable_shapes",  # min_dims is invalid type
            num_shapes=2,
            min_dims="a",
        ),
        e(
            "mutually_broadcastable_shapes",  # max_side is invalid type
            num_shapes=2,
            max_side="a",
        ),
        e(
            "mutually_broadcastable_shapes",  # max_side is invalid type
            num_shapes=2,
            max_dims="a",
        ),
        e(
            "mutually_broadcastable_shapes",  # min_side is out of domain
            num_shapes=2,
            min_side=-1,
        ),
        e(
            "mutually_broadcastable_shapes",  # min_dims is out of domain
            num_shapes=2,
            min_dims=-1,
        ),
        e(
            "mutually_broadcastable_shapes",  # max_side < min_side
            num_shapes=2,
            min_side=1,
            max_side=0,
        ),
        e(
            "mutually_broadcastable_shapes",  # max_dims < min_dims
            num_shapes=2,
            min_dims=1,
            max_dims=0,
        ),
        e(
            "mutually_broadcastable_shapes",  # max_side too small
            num_shapes=2,
            base_shape=(5, 1),
            min_dims=2,
            max_dims=4,
            min_side=2,
            max_side=3,
        ),
        e(
            "mutually_broadcastable_shapes",  # min_side too large
            num_shapes=2,
            base_shape=(0, 1),
            min_dims=2,
            max_dims=4,
            min_side=2,
            max_side=3,
        ),
        e(
            "mutually_broadcastable_shapes",  # user-specified max_dims unsatisfiable
            num_shapes=1,
            base_shape=(5, 3, 2, 1),
            min_dims=3,
            max_dims=4,
            min_side=2,
            max_side=3,
        ),
        e(
            "mutually_broadcastable_shapes",  # user-specified max_dims unsatisfiable
            num_shapes=2,
            base_shape=(0, 3, 2, 1),
            min_dims=3,
            max_dims=4,
            min_side=2,
            max_side=3,
        ),
        e("indices", shape=0),
        e("indices", shape=("1", "2")),
        e("indices", shape=(0, -1)),
        e("indices", shape=(0, 0), allow_newaxis=None),
        e("indices", shape=(0, 0), allow_ellipsis=None),
        e("indices", shape=(0, 0), min_dims=-1),
        e("indices", shape=(0, 0), min_dims=1.0),
        e("indices", shape=(0, 0), max_dims=-1),
        e("indices", shape=(0, 0), max_dims=1.0),
        e("indices", shape=(0, 0), min_dims=2, max_dims=1),
        e("indices", shape=(3, 3, 3), min_dims=4),
        e("indices", shape=(3, 3, 3), max_dims=5),
        e("indices", shape=5, min_dims=0),
        e("indices", shape=(5,), min_dims=2),
        e("indices", shape=(5,), max_dims=2),
    ],
)
def test_raise_invalid_argument(xp, xps, strat_name, kwargs):
    """Strategies raise helpful error with invalid arguments."""
    strat_func = getattr(xps, strat_name)
    strat = strat_func(**kwargs)
    with pytest.raises(InvalidArgument):
        check_can_generate_examples(strat)


@pytest.mark.parametrize("api_version", [..., "latest", "1970.01", 42])
def test_make_strategies_namespace_raise_invalid_argument(xp, api_version):
    """Function raises helpful error with invalid arguments."""
    with pytest.raises(InvalidArgument):
        make_strategies_namespace(xp, api_version=api_version)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import os
import sys
import tempfile
import time
import unicodedata
from typing import get_args

import pytest

from hypothesis import given, strategies as st
from hypothesis.errors import InvalidArgument
from hypothesis.internal import charmap as cm
from hypothesis.internal.intervalsets import IntervalSet

from tests.common.utils import skipif_threading


def test_charmap_contains_all_unicode():
    n = 0
    for vs in cm.charmap().values():
        for u, v in vs:
            n += v - u + 1
    assert n == sys.maxunicode + 1


def test_charmap_has_right_categories():
    for cat, intervals in cm.charmap().items():
        for u, v in intervals:
            for i in range(u, v + 1):
                real = unicodedata.category(chr(i))
                assert real == cat, f"{i} is {real} but reported in {cat}"


def assert_valid_range_list(ls):
    for u, v in ls:
        assert u <= v
    for i in range(len(ls) - 1):
        assert ls[i] <= ls[i + 1]
        assert ls[i][-1] < ls[i + 1][0]


@given(st.sets(st.sampled_from(cm.categories())))
def test_query_matches_categories(cats):
    values = cm.query(categories=cats).intervals
    assert_valid_range_list(values)
    for u, v in values:
        for i in (u, v, (u + v) // 2):
            assert unicodedata.category(chr(i)) in cats


@given(
    st.sets(st.sampled_from(cm.categories())) | st.none(),
    st.integers(0, sys.maxunicode),
    st.integers(0, sys.maxunicode),
)
def test_query_matches_categories_codepoints(cats, m1, m2):
    m1, m2 = sorted((m1, m2))
    values = cm.query(categories=cats, min_codepoint=m1, max_codepoint=m2).intervals
    assert_valid_range_list(values)
    for u, v in values:
        assert m1 <= u
        assert v <= m2


# any test which sets `_charmap = None` is thread-unsafe.
@skipif_threading
def test_reload_charmap():
    x = cm.charmap()
    assert x is cm.charmap()
    cm._charmap = None
    y = cm.charmap()
    assert x is not y
    assert x == y


@skipif_threading
def test_recreate_charmap():
    x = cm.charmap()
    assert x is cm.charmap()
    cm._charmap = None
    cm.charmap_file().unlink()
    y = cm.charmap()
    assert x is not y
    assert x == y


# This test fails flakily (every 1 in ~10 full CI runs), but only on the test-pyodide
# ci job:
#
#      os.utime(cm.charmap_file(), (mtime, mtime))
#      ~~~~~~~~^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
#   FileNotFoundError: [Errno 44] No such file or directory:
#   '/home/runner/work/hypothesis/hypothesis/.hypothesis/unicode_data/15.1.0/charmap.json.gz'
#
# I suspect this is from a race condition due to how we parallelize the pyodide
# tests in CI, (splitting test files across 20 processes). It's also possible
# it's from some pyodide-specific weirdness, but since the test only fails sometimes,
# I find this less likely.
#
# I'm xfailing this on emscripten for now, to get a consistent green CI.
@pytest.mark.xfail(
    condition=sys.platform == "emscripten", strict=False, reason="see comment"
)
@skipif_threading
def test_uses_cached_charmap():
    cm.charmap()

    # Reset the last-modified time of the cache file to a point in the past.
    mtime = int(time.time() - 1000)
    os.utime(cm.charmap_file(), (mtime, mtime))
    statinfo = cm.charmap_file().stat()
    assert statinfo.st_mtime == mtime

    # Force reload of charmap from cache file and check that mtime is unchanged.
    cm._charmap = None
    cm.charmap()
    statinfo = cm.charmap_file().stat()
    assert statinfo.st_mtime == mtime


def _union_intervals(x, y):
    return IntervalSet(x).union(IntervalSet(y)).intervals


def test_union_empty():
    assert _union_intervals([], []) == ()
    assert _union_intervals([], [[1, 2]]) == ((1, 2),)
    assert _union_intervals([[1, 2]], []) == ((1, 2),)


def test_union_handles_totally_overlapped_gap():
    #   < xx  >  Imagine the intervals x and y as bit strings.
    # | <yy yy>  The bit at position n is set if n falls inside that interval.
    # = <zzzzz>  In this model _union_intervals() performs bit-wise or.
    assert _union_intervals([[2, 3]], [[1, 2], [4, 5]]) == ((1, 5),)


def test_union_handles_partially_overlapped_gap():
    #   <  x  >  Imagine the intervals x and y as bit strings.
    # | <yy  y>  The bit at position n is set if n falls inside that interval.
    # = <zzz z>  In this model _union_intervals() performs bit-wise or.
    assert _union_intervals([[3, 3]], [[1, 2], [5, 5]]) == ((1, 3), (5, 5))


def test_successive_union():
    x = []
    for v in cm.charmap().values():
        x = _union_intervals(x, v)
    assert x == ((0, sys.maxunicode),)


@skipif_threading
def test_can_handle_race_between_exist_and_create(monkeypatch):
    x = cm.charmap()
    cm._charmap = None
    monkeypatch.setattr(os.path, "exists", lambda p: False)
    y = cm.charmap()
    assert x is not y
    assert x == y


@skipif_threading
def test_exception_in_write_does_not_lead_to_broken_charmap(monkeypatch):
    def broken(*args, **kwargs):
        raise ValueError

    cm._charmap = None
    monkeypatch.setattr(os.path, "exists", lambda p: False)
    monkeypatch.setattr(os, "rename", broken)

    cm.charmap()
    cm.charmap()


@skipif_threading
def test_regenerate_broken_charmap_file():
    cm.charmap()

    cm.charmap_file().write_bytes(b"")  # overwrite with empty file

    cm._charmap = None
    cm.charmap()


def test_query_rejects_min_codepoint_greater_than_max():
    with pytest.raises(InvalidArgument):
        cm.query(min_codepoint=1, max_codepoint=0)


def test_exclude_characters_are_included_in_key():
    assert cm.query().intervals != cm.query(exclude_characters="0").intervals


@skipif_threading
def test_error_writing_charmap_file_is_suppressed(monkeypatch):
    def broken_mkstemp(dir):
        raise RuntimeError

    monkeypatch.setattr(tempfile, "mkstemp", broken_mkstemp)

    try:
        # Cache the charmap to avoid a performance hit the next time
        # somebody tries to use it.
        saved = cm._charmap
        cm._charmap = None
        cm.charmap_file().unlink()

        cm.charmap()
    finally:
        cm._charmap = saved


def test_categoryname_literal_is_correct():
    minor_categories = set(cm.categories())
    major_categories = {c[0] for c in minor_categories}
    assert set(get_args(cm.CategoryName)) == minor_categories | major_categories



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import pytest

from hypothesis import given, settings, strategies as st
from hypothesis.database import InMemoryExampleDatabase
from hypothesis.internal.compat import ExceptionGroup


def test_does_not_shrink_on_replay():
    database = InMemoryExampleDatabase()

    call_count = 0

    is_first = True
    last = None

    @settings(
        database=database,
        report_multiple_bugs=False,
        derandomize=False,
        max_examples=1000,
    )
    @given(st.lists(st.integers(), unique=True, min_size=3))
    def test(ls):
        nonlocal call_count, is_first, last
        if is_first and last is not None:
            assert ls == last
        is_first = False
        last = ls
        call_count += 1
        raise AssertionError

    with pytest.raises(AssertionError):
        test()

    assert last is not None

    call_count = 0
    is_first = True

    with pytest.raises(AssertionError):
        test()

    assert call_count == 2


def test_does_not_shrink_on_replay_with_multiple_bugs():
    database = InMemoryExampleDatabase()
    call_count = 0
    raised = False
    marker = 1000093

    @settings(
        database=database,
        report_multiple_bugs=True,
        derandomize=False,
        max_examples=1000,
    )
    @given(st.integers())
    def test(n):
        nonlocal call_count, raised
        call_count += 1
        if n >= marker:
            raised = True
            raise AssertionError
        elif n < marker and raised:
            raise AssertionError

    with pytest.raises(ExceptionGroup):
        test()

    call_count = 0

    with pytest.raises(ExceptionGroup):
        test()

    assert call_count == 4


def test_will_always_shrink_if_previous_example_does_not_replay():
    database = InMemoryExampleDatabase()

    good = set()
    last = None

    @settings(
        database=database,
        report_multiple_bugs=True,
        derandomize=False,
        max_examples=1000,
    )
    @given(st.integers(min_value=0))
    def test(n):
        nonlocal last
        if n not in good:
            last = n
            raise AssertionError

    for i in range(20):
        with pytest.raises(AssertionError):
            test()
        assert last == i
        good.add(last)


def test_will_shrink_if_the_previous_example_does_not_look_right():
    database = InMemoryExampleDatabase()

    last = None

    first_test = True

    @settings(database=database, report_multiple_bugs=True, derandomize=False)
    @given(st.data())
    def test(data):
        nonlocal last
        m = data.draw(st.integers())
        last = m
        if first_test:
            data.draw(st.integers())
            assert m < 10000
        else:
            raise AssertionError

    with pytest.raises(AssertionError):
        test()

    assert last is not None
    assert last > 0

    first_test = False

    with pytest.raises(AssertionError):
        test()

    assert last == 0



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import pytest

from hypothesis import given, strategies as st
from hypothesis.strategies import integers, lists


@pytest.mark.parametrize(
    ("specifier", "condition"),
    [(integers(), lambda x: x > 1), (lists(integers()), bool)],
)
def test_filter_correctly(specifier, condition):
    @given(specifier.filter(condition))
    def test_is_filtered(x):
        assert condition(x)

    test_is_filtered()


# A variety of strategies that generate the integers 1-20 inclusive, but might
# differ in their support for special-case filtering.
one_to_twenty_strategies = [
    st.integers(1, 20),
    st.integers(0, 19).map(lambda x: x + 1),
    st.sampled_from(range(1, 21)),
    st.sampled_from(range(20)).map(lambda x: x + 1),
]


@pytest.mark.parametrize("base", one_to_twenty_strategies)
@given(
    data=st.data(),
    forbidden_values=st.lists(st.integers(1, 20), max_size=19, unique=True),
)
def test_chained_filters_agree(data, forbidden_values, base):
    def forbid(s, forbidden):
        """Helper function to avoid Python variable scoping issues."""
        return s.filter(lambda x: x != forbidden)

    s = base
    for forbidden in forbidden_values:
        s = forbid(s, forbidden)

    x = data.draw(s)
    assert 1 <= x <= 20
    assert x not in forbidden_values


@pytest.mark.parametrize("base", one_to_twenty_strategies)
def test_chained_filters_repr(base):
    def foo(x):
        return x != 0

    def bar(x):
        return x != 2

    filtered = base.filter(foo)
    chained = filtered.filter(bar)
    assert repr(chained) == f"{base!r}.filter(foo).filter(bar)"
    assert repr(filtered) == f"{base!r}.filter(foo)"



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import re
from typing import NamedTuple

from hypothesistooling.__main__ import PYTHONS as pythons_map

PYTHON_VERSIONS = [v for v in pythons_map if re.fullmatch(r"3\.\d\d?", v)]

try:
    from numpy import __version__ as np_version
except ImportError:
    NP1 = False
else:
    NP1 = np_version.startswith("1.")

ASSUME_REVEALED_TYPES = [
    ("assume(False)", "Never"),
    ("assume(None)", "Never"),
    ("assume(True)", "Literal[True]"),
    ("assume(1)", "Literal[True]"),
]

REVEALED_TYPES = [
    ("integers()", "int"),
    ("text()", "str"),
    ("integers().map(str)", "str"),
    ("booleans().filter(bool)", "bool"),
    ("tuples()", "tuple[()]"),
    ("tuples(integers())", "tuple[int]"),
    ("tuples(integers(), text())", "tuple[int, str]"),
    (
        "tuples(integers(), text(), integers(), text(), integers())",
        "tuple[int, str, int, str, int]",
    ),
    (
        "tuples(text(), text(), text(), text(), text(), text())",
        "tuple[Any, ...]",
    ),
    ("lists(none())", "list[None]"),
    ("integers().filter(lambda x: x > 0)", "int"),
    ("booleans().filter(lambda x: x)", "bool"),
    ("integers().map(bool).filter(lambda x: x)", "bool"),
    ("integers().flatmap(lambda x: lists(floats()))", "list[float]"),
    ("one_of([integers(), integers()])", "int"),
    ("one_of([integers(), floats()])", "float"),
    ("one_of([integers(), none()])", "int | None"),
    ("one_of(integers(), none())", "int | None"),
    ("one_of(integers(), text())", "int | str"),
    ("recursive(integers(), lists)", "list[Any] | int"),
    # We have overloads for up to five types, then fall back to Any.
    # (why five?  JSON atoms are None|bool|int|float|str and we do that a lot)
    (
        "one_of(integers(), text(), none(), binary(), builds(list), builds(dict))",
        "Any",
    ),
    ("from_regex(r'.', alphabet=None)", "str"),
]


class DifferingRevealedTypes(NamedTuple):
    value: str
    mypy: str
    pyright: str


DIFF_REVEALED_TYPES = [
    DifferingRevealedTypes("none() | integers()", "None | int", "int | None"),
    DifferingRevealedTypes(
        "data()", "hypothesis.strategies._internal.core.DataObject", "DataObject"
    ),
    # We have overloads for up to five types, then fall back to Any.
    # (why five?  JSON atoms are None|bool|int|float|str and we do that a lot)
    DifferingRevealedTypes(
        "one_of(integers(), text(), none(), binary(), builds(list))",
        "int | str | None | bytes | list[Never]",
        "int | str | bytes | list[Unknown] | None",
    ),
    DifferingRevealedTypes(
        "dictionaries(integers(), datetimes())",
        "dict[int, datetime.datetime]",
        "dict[int, datetime]",
    ),
]


NUMPY_REVEALED_TYPES = [
    (
        'arrays(dtype=np.dtype("int32"), shape=1)',
        "ndarray[tuple[Any, ...], dtype[signedinteger[_32Bit]]]",
    ),
    # (
    #     "arrays(dtype=np.dtype(int), shape=1)",
    #     "ndarray[tuple[int, ...], dtype[Union[signedinteger[Union[_32Bit, _64Bit]], bool[bool]]]]",
    #     # FIXME: `dtype[signedinteger[_32Bit | _64Bit] | bool[bool]]]]` on mypy now
    # ),
    (
        "boolean_dtypes()",
        "dtype[bool[bool]]",  # np.bool[builtins.bool]
    ),
    (
        "unsigned_integer_dtypes(sizes=8)",
        "dtype[unsignedinteger[_8Bit]]",
    ),
    (
        "unsigned_integer_dtypes(sizes=16)",
        "dtype[unsignedinteger[_16Bit]]",
    ),
    (
        "unsigned_integer_dtypes(sizes=32)",
        "dtype[unsignedinteger[_32Bit]]",
    ),
    (
        "unsigned_integer_dtypes(sizes=64)",
        "dtype[unsignedinteger[_64Bit]]",
    ),
    (
        "unsigned_integer_dtypes()",
        "dtype[unsignedinteger[Any]]",
    ),
    (
        "unsigned_integer_dtypes(sizes=(8, 16))",
        "dtype[unsignedinteger[Any]]",
    ),
    (
        "integer_dtypes(sizes=8)",
        "dtype[signedinteger[_8Bit]]",
    ),
    (
        "integer_dtypes(sizes=16)",
        "dtype[signedinteger[_16Bit]]",
    ),
    (
        "integer_dtypes(sizes=32)",
        "dtype[signedinteger[_32Bit]]",
    ),
    (
        "integer_dtypes(sizes=64)",
        "dtype[signedinteger[_64Bit]]",
    ),
    (
        "integer_dtypes()",
        "dtype[signedinteger[Any]]",
    ),
    (
        "integer_dtypes(sizes=(8, 16))",
        "dtype[signedinteger[Any]]",
    ),
    (
        "floating_dtypes(sizes=16)",
        "dtype[floating[_16Bit]]",
    ),
    (
        "floating_dtypes(sizes=32)",
        "dtype[floating[_32Bit]]",
    ),
    (
        "floating_dtypes(sizes=64)",
        "dtype[float64]",
    ),
    (
        "floating_dtypes(sizes=128)",
        "dtype[floating[_128Bit]]",
    ),
    (
        "floating_dtypes()",
        "dtype[floating[Any]]",
    ),
    (
        "floating_dtypes(sizes=(16, 32))",
        "dtype[floating[Any]]",
    ),
    (
        "complex_number_dtypes(sizes=64)",
        "dtype[complexfloating[_32Bit, _32Bit]]",
    ),
    (
        "complex_number_dtypes(sizes=128)",
        "dtype[complex128]",
    ),
    (
        "complex_number_dtypes(sizes=256)",
        "dtype[complexfloating[_128Bit, _128Bit]]",
    ),
    (
        "complex_number_dtypes()",
        "dtype[complexfloating[Any, Any]]",
    ),
    (
        "complex_number_dtypes(sizes=(64, 128))",
        "dtype[complexfloating[Any, Any]]",
    ),
    (
        "integer_array_indices(shape=(2, 3))",
        "tuple[ndarray[tuple[Any, ...], dtype[signedinteger[Any]]], ...]",
    ),
    (
        'integer_array_indices(shape=(2, 3), dtype=np.dtype("int32"))',
        "tuple[ndarray[tuple[Any, ...], dtype[signedinteger[_32Bit]]], ...]",
    ),
    (
        'integer_array_indices(shape=(2, 3), dtype=np.dtype("uint8"))',
        "tuple[ndarray[tuple[Any, ...], dtype[unsignedinteger[_8Bit]]], ...]",
    ),
    # basic_indices with allow_ellipsis=False (no EllipsisType differences)
    (
        "basic_indices((3, 4), allow_ellipsis=False)",
        "int | slice[Any, Any, Any] | tuple[int | slice[Any, Any, Any], ...]",
    ),
]

# basic_indices tests where mypy/pyright differ in EllipsisType representation
NUMPY_DIFF_REVEALED_TYPES = [
    # mypy uses types.EllipsisType, pyright uses EllipsisType
    DifferingRevealedTypes(
        "basic_indices((3, 4))",
        "int | slice[Any, Any, Any] | types.EllipsisType | tuple[int | slice[Any, Any, Any] | types.EllipsisType, ...]",
        "int | slice[Any, Any, Any] | EllipsisType | tuple[int | slice[Any, Any, Any] | EllipsisType, ...]",
    ),
    # pyright also reorders None to the end
    DifferingRevealedTypes(
        "basic_indices((3, 4), allow_newaxis=True, allow_ellipsis=False)",
        "int | slice[Any, Any, Any] | None | tuple[int | slice[Any, Any, Any] | None, ...]",
        "int | slice[Any, Any, Any] | tuple[int | slice[Any, Any, Any] | None, ...] | None",
    ),
    DifferingRevealedTypes(
        "basic_indices((3, 4), allow_newaxis=True)",
        "int | slice[Any, Any, Any] | None | types.EllipsisType | tuple[int | slice[Any, Any, Any] | None | types.EllipsisType, ...]",
        "int | slice[Any, Any, Any] | EllipsisType | tuple[int | slice[Any, Any, Any] | EllipsisType | None, ...] | None",
    ),
]



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import traceback

import pytest

from hypothesis import Verbosity, given, settings, strategies as st


@pytest.mark.parametrize("verbosity", [Verbosity.normal, Verbosity.debug])
@pytest.mark.parametrize("env_value", ["", "1"])
def test_tracebacks_omit_hypothesis_internals(monkeypatch, env_value, verbosity):
    monkeypatch.setenv("HYPOTHESIS_NO_TRACEBACK_TRIM", env_value)

    @settings(verbosity=verbosity)
    @given(st.just(False))
    def simplest_failure(x):
        raise ValueError

    try:
        simplest_failure()
    except ValueError as e:
        tb = traceback.extract_tb(e.__traceback__)
        # Unless in debug mode, Hypothesis adds 1 frame - the least possible!
        # (4 frames: this one, simplest_failure, internal frame, assert False)
        if verbosity < Verbosity.debug and not env_value:
            assert len(tb) == 4
        else:
            assert len(tb) >= 5



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from hypothesis import given, strategies as st

from tests.common.utils import run_test_for_falsifying_example
from tests.snapshots.conftest import EXPLAIN_SETTINGS


def test_explain_comments_basic_fail_if_either(snapshot):
    @EXPLAIN_SETTINGS
    @given(
        st.booleans(),
        st.booleans(),
        st.lists(st.none()),
        st.booleans(),
        st.booleans(),
    )
    def inner(a, b, c, d, e):
        assert not (b and d)

    assert run_test_for_falsifying_example(inner) == snapshot


def test_explain_comments_basic_fail_if_not_all(snapshot):
    @EXPLAIN_SETTINGS
    @given(st.text(), st.text(), st.text())
    def inner(a, b, c):
        condition = a and b and c
        assert condition

    assert run_test_for_falsifying_example(inner) == snapshot


def test_explain_no_together_comment_if_single_argument(snapshot):
    @EXPLAIN_SETTINGS
    @given(st.text(), st.text())
    def inner(a, b):
        assert a

    assert run_test_for_falsifying_example(inner) == snapshot


class MyClass:
    def __init__(self, x, y):
        self.x = x
        self.y = y


def test_explain_builds_subargs(snapshot):
    @EXPLAIN_SETTINGS
    @given(st.builds(MyClass, st.integers(), st.booleans()))
    def inner(obj):
        assert not obj.y

    assert run_test_for_falsifying_example(inner) == snapshot


def test_explain_builds_kwargs_subargs(snapshot):
    @EXPLAIN_SETTINGS
    @given(st.builds(MyClass, x=st.integers(), y=st.booleans()))
    def inner(obj):
        assert not obj.y

    assert run_test_for_falsifying_example(inner) == snapshot


def test_explain_tuple_subargs(snapshot):
    @EXPLAIN_SETTINGS
    @given(st.tuples(st.integers(), st.booleans()))
    def inner(t):
        assert not t[1]

    assert run_test_for_falsifying_example(inner) == snapshot


def test_explain_fixeddict_subargs(snapshot):
    @EXPLAIN_SETTINGS
    @given(st.fixed_dictionaries({"x": st.integers(), "y": st.booleans()}))
    def inner(d):
        assert not d["y"]

    assert run_test_for_falsifying_example(inner) == snapshot


def test_explain_tuple_multiple_varying(snapshot):
    @EXPLAIN_SETTINGS
    @given(st.tuples(st.integers(), st.text(), st.booleans()))
    def inner(t):
        assert not t[2]

    assert run_test_for_falsifying_example(inner) == snapshot


def test_explain_skip_subset_slices(snapshot):
    @EXPLAIN_SETTINGS
    @given(st.builds(MyClass, st.tuples(st.integers(), st.booleans()), y=st.booleans()))
    def inner(obj):
        assert obj.y

    assert run_test_for_falsifying_example(inner) == snapshot


def test_explain_duplicate_param_names(snapshot):
    @EXPLAIN_SETTINGS
    @given(
        kw=st.integers(),
        b=st.fixed_dictionaries({"kw": st.text(), "c": st.booleans()}),
    )
    def inner(kw, b):
        assert not b["c"]

    assert run_test_for_falsifying_example(inner) == snapshot


class Outer:
    def __init__(self, inner, value):
        self.inner = inner
        self.value = value


class Inner:
    def __init__(self, x):
        self.x = x


def test_explain_multi_level_nesting(snapshot):
    @EXPLAIN_SETTINGS
    @given(
        bare=st.integers(),
        outer=st.builds(
            Outer, inner=st.builds(Inner, x=st.integers()), value=st.booleans()
        ),
    )
    def inner(bare, outer):
        assert not outer.value

    assert run_test_for_falsifying_example(inner) == snapshot



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import os
import subprocess
import sys

TESTSUITE = """
from hypothesis import given
from hypothesis import strategies as st
import _pytest

@given(st.integers())
def test_a(n):
    pass

test_a()
"""


def test_import_just_private_pytest_works(testdir):
    # importing just _pytest does not import its submodules like _pytest.outcomes.
    # Ensure we don't rely on this by writing conditional imports like
    # `if "_pytest" in sys.modules: sys.modules["_pytest"].outcomes`.
    script = testdir.makepyfile(TESTSUITE)
    result = subprocess.run(
        [sys.executable, script],
        # some hypothesis plugins (trio) import pytest, making this test pass
        # even when hypothesis has a bug. This was the case for my local installation,
        # and I am including this here on the off chance our CI setup has a similar
        # install situation. I don't want to wait for another regression to find out
        # this test doesn't work.
        env=os.environ | {"HYPOTHESIS_NO_PLUGINS": "1"},
    )
    assert result.returncode == 0



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import re
from collections import defaultdict
from typing import ClassVar

import pytest
from _pytest.outcomes import Failed, Skipped
from pytest import raises

from hypothesis import (
    HealthCheck,
    Phase,
    __version__,
    reproduce_failure,
    seed,
    settings as Settings,
    strategies as st,
)
from hypothesis.control import current_build_context
from hypothesis.core import encode_failure
from hypothesis.database import InMemoryExampleDatabase
from hypothesis.errors import (
    DidNotReproduce,
    Flaky,
    FlakyStrategyDefinition,
    InvalidArgument,
    InvalidDefinition,
)
from hypothesis.stateful import (
    Bundle,
    RuleBasedStateMachine,
    consumes,
    get_state_machine_test,
    initialize,
    invariant,
    multiple,
    precondition,
    rule,
    run_state_machine_as_test,
)
from hypothesis.strategies import binary, data, integers, just, lists

from tests.common.utils import (
    Why,
    capture_out,
    skipif_threading,
    validate_deprecation,
    xfail_on_crosshair,
)
from tests.nocover.test_stateful import DepthMachine

NO_BLOB_SETTINGS = Settings(print_blob=False, phases=tuple(Phase)[:-1])


class MultipleRulesSameFuncMachine(RuleBasedStateMachine):
    def myfunc(self, data):
        print(data)

    rule1 = rule(data=just("rule1data"))(myfunc)
    rule2 = rule(data=just("rule2data"))(myfunc)


class PreconditionMachine(RuleBasedStateMachine):
    num = 0

    @rule()
    def add_one(self):
        self.num += 1

    @rule()
    def set_to_zero(self):
        self.num = 0

    @rule(num=integers())
    @precondition(lambda self: self.num != 0)
    def div_by_precondition_after(self, num):
        self.num = num / self.num

    @precondition(lambda self: self.num != 0)
    @rule(num=integers())
    def div_by_precondition_before(self, num):
        self.num = num / self.num


TestPrecondition = PreconditionMachine.TestCase
TestPrecondition.settings = Settings(TestPrecondition.settings, max_examples=10)


def test_picks_up_settings_at_first_use_of_testcase():
    assert TestPrecondition.settings.max_examples == 10


def test_multiple_rules_same_func():
    test_class = MultipleRulesSameFuncMachine.TestCase
    with capture_out() as o:
        test_class().runTest()
    output = o.getvalue()
    assert "rule1data" in output
    assert "rule2data" in output


def test_can_get_test_case_off_machine_instance():
    assert DepthMachine().TestCase is DepthMachine().TestCase
    assert DepthMachine().TestCase is not None


class FlakyDrawLessMachine(RuleBasedStateMachine):
    @rule(d=data())
    def action(self, d):
        if current_build_context().is_final:
            d.draw(binary(min_size=1, max_size=1))
        else:
            buffer = binary(min_size=1024, max_size=1024)
            assert 0 not in buffer


def test_flaky_draw_less_raises_flaky():
    with raises(Flaky):
        FlakyDrawLessMachine.TestCase().runTest()


def test_result_is_added_to_target():
    class TargetStateMachine(RuleBasedStateMachine):
        nodes = Bundle("nodes")

        @rule(target=nodes, source=lists(nodes))
        def bunch(self, source):
            assert len(source) == 0
            return source

    test_class = TargetStateMachine.TestCase
    try:
        test_class().runTest()
        raise RuntimeError("Expected an assertion error")
    except AssertionError as err:
        notes = err.__notes__
    regularized_notes = [re.sub(r"[0-9]+", "i", note) for note in notes]
    assert "state.bunch(source=[nodes_i])" in regularized_notes


class FlakyStateMachine(RuleBasedStateMachine):
    @rule()
    def action(self):
        assert current_build_context().is_final


def test_flaky_raises_flaky():
    with raises(Flaky):
        FlakyStateMachine.TestCase().runTest()


class FlakyPreconditionMachine(RuleBasedStateMachine):
    @precondition(lambda self: not current_build_context().is_final)
    @rule()
    def action(self):
        raise AssertionError


def test_flaky_precondition_error_message():
    with raises(FlakyStrategyDefinition) as exc_info:
        FlakyPreconditionMachine.TestCase().runTest()
    assert any("flaky precondition" in note for note in exc_info.value.__notes__)


class FlakyDrawInRuleMachine(RuleBasedStateMachine):
    # Flakiness inside rule execution (via data().draw()) happens AFTER rule selection,
    # so the "flaky precondition" note should NOT be added.
    @rule(d=data())
    def action(self, d):
        if current_build_context().is_final:
            d.draw(st.integers(0, 0))
        d.draw(st.integers())
        raise AssertionError


def test_flaky_draw_in_rule_no_precondition_note():
    # When flakiness occurs during rule execution (not rule selection),
    # the error message should NOT mention flaky preconditions.
    with raises(FlakyStrategyDefinition) as exc_info:
        FlakyDrawInRuleMachine.TestCase().runTest()
    notes = getattr(exc_info.value, "__notes__", [])
    assert not any("flaky precondition" in note for note in notes)


def test_get_state_machine_test_is_importable():
    # Regression test: get_state_machine_test is used by HypoFuzz
    assert callable(get_state_machine_test)


class FlakyRatchettingMachine(RuleBasedStateMachine):
    ratchet = 0

    @rule(d=data())
    def action(self, d):
        FlakyRatchettingMachine.ratchet += 1
        n = FlakyRatchettingMachine.ratchet
        d.draw(lists(integers(), min_size=n, max_size=n))
        raise AssertionError


@Settings(
    stateful_step_count=10,
    max_examples=30,
    suppress_health_check=[HealthCheck.filter_too_much],
)  # speed this up
class MachineWithConsumingRule(RuleBasedStateMachine):
    b1 = Bundle("b1")
    b2 = Bundle("b2")

    def __init__(self):
        self.created_counter = 0
        self.consumed_counter = 0
        super().__init__()

    @invariant()
    def bundle_length(self):
        assert len(self.bundle("b1")) == self.created_counter - self.consumed_counter

    @rule(target=b1)
    def populate_b1(self):
        self.created_counter += 1
        return self.created_counter

    @rule(target=b2, consumed=consumes(b1))
    def depopulate_b1(self, consumed):
        self.consumed_counter += 1
        return consumed

    @rule(consumed=lists(consumes(b1), max_size=3))
    def depopulate_b1_multiple(self, consumed):
        self.consumed_counter += len(consumed)

    @rule(value1=b1, value2=b2)
    def check(self, value1, value2):
        assert value1 != value2


TestMachineWithConsumingRule = MachineWithConsumingRule.TestCase


def test_multiple():
    none = multiple()
    some = multiple(1, 2.01, "3", b"4", 5)
    assert len(none.values) == 0
    assert len(some.values) == 5
    assert set(some.values) == {1, 2.01, "3", b"4", 5}


class MachineUsingMultiple(RuleBasedStateMachine):
    b = Bundle("b")

    def __init__(self):
        self.expected_bundle_length = 0
        super().__init__()

    @invariant()
    def bundle_length(self):
        assert len(self.bundle("b")) == self.expected_bundle_length

    @rule(target=b, items=lists(elements=integers(), max_size=10))
    def populate_bundle(self, items):
        self.expected_bundle_length += len(items)
        return multiple(*items)

    @rule(target=b)
    def do_not_populate(self):
        return multiple()


TestMachineUsingMultiple = MachineUsingMultiple.TestCase


def test_multiple_variables_printed():
    class ProducesMultiple(RuleBasedStateMachine):
        b = Bundle("b")

        @initialize(target=b)
        def populate_bundle(self):
            return multiple(1, 2)

        @rule()
        def fail_fast(self):
            raise AssertionError

    with raises(AssertionError) as err:
        run_state_machine_as_test(ProducesMultiple)

    # This is tightly coupled to the output format of the step printing.
    # The first line is "Falsifying Example:..." the second is creating
    # the state machine, the third is calling the "initialize" method.
    assignment_line = err.value.__notes__[2]
    # 'populate_bundle()' returns 2 values, so should be
    # expanded to 2 variables.
    assert assignment_line == "b_0, b_1 = state.populate_bundle()"

    # Make sure MultipleResult is iterable so the printed code is valid.
    # See https://github.com/HypothesisWorks/hypothesis/issues/2311
    state = ProducesMultiple()
    _b_0, _b_1 = state.populate_bundle()
    with raises(AssertionError):
        state.fail_fast()


def test_multiple_variables_printed_single_element():
    # https://github.com/HypothesisWorks/hypothesis/issues/3236
    class ProducesMultiple(RuleBasedStateMachine):
        b = Bundle("b")

        @initialize(target=b)
        def populate_bundle(self):
            return multiple(1)

        @rule(b=b)
        def fail_fast(self, b):
            assert b != 1

    with raises(AssertionError) as err:
        run_state_machine_as_test(ProducesMultiple)

    assignment_line = err.value.__notes__[2]
    assert assignment_line == "(b_0,) = state.populate_bundle()"

    state = ProducesMultiple()
    (v1,) = state.populate_bundle()
    state.fail_fast((v1,))  # passes if tuple not unpacked
    with raises(AssertionError):
        state.fail_fast(v1)


def test_no_variables_printed():
    class ProducesNoVariables(RuleBasedStateMachine):
        b = Bundle("b")

        @initialize(target=b)
        def populate_bundle(self):
            return multiple()

        @rule()
        def fail_fast(self):
            raise AssertionError

    with raises(AssertionError) as err:
        run_state_machine_as_test(ProducesNoVariables)

    # This is tightly coupled to the output format of the step printing.
    # The first line is "Falsifying Example:..." the second is creating
    # the state machine, the third is calling the "initialize" method.
    assignment_line = err.value.__notes__[2]
    # 'populate_bundle()' returns 0 values, so there should be no
    # variable assignment.
    assert assignment_line == "state.populate_bundle()"


def test_consumes_typecheck():
    with pytest.raises(TypeError):
        consumes(integers())


def test_ratchetting_raises_flaky():
    with raises(Flaky):
        FlakyRatchettingMachine.TestCase().runTest()


def test_empty_machine_is_invalid():
    class EmptyMachine(RuleBasedStateMachine):
        pass

    with raises(InvalidDefinition):
        EmptyMachine.TestCase().runTest()


def test_machine_with_no_terminals_is_invalid():
    class NonTerminalMachine(RuleBasedStateMachine):
        @rule(value=Bundle("hi"))
        def bye(self, hi):
            pass

    with raises(InvalidDefinition):
        NonTerminalMachine.TestCase().runTest()


def test_minimizes_errors_in_teardown():
    # temporary debugging to try to narrow down a potential thread-safety issue
    import threading

    from hypothesis import Verbosity

    counter = 0

    @Settings(database=None, verbosity=Verbosity.debug)
    class Foo(RuleBasedStateMachine):
        @initialize()
        def init(self):
            nonlocal counter
            counter = 0
            print(f"[{threading.get_ident()}] init", counter)

        @rule()
        def increment(self):
            nonlocal counter
            counter += 1
            print(f"[{threading.get_ident()}] increment", counter)

        def teardown(self):
            nonlocal counter
            print(f"[{threading.get_ident()}] teardown", counter)
            assert not counter

    with raises(AssertionError):
        run_state_machine_as_test(Foo)
    assert counter == 1


class RequiresInit(RuleBasedStateMachine):
    def __init__(self, threshold):
        super().__init__()
        self.threshold = threshold

    @rule(value=integers())
    def action(self, value):
        if value > self.threshold:
            raise ValueError(f"{value} is too high")


def test_can_use_factory_for_tests():
    with raises(ValueError):
        run_state_machine_as_test(
            lambda: RequiresInit(42), settings=Settings(max_examples=100)
        )


class FailsEventually(RuleBasedStateMachine):
    def __init__(self):
        super().__init__()
        self.counter = 0

    @rule()
    def increment(self):
        self.counter += 1
        assert self.counter < 10


FailsEventually.TestCase.settings = Settings(stateful_step_count=5)


@skipif_threading
def test_can_explicitly_pass_settings():
    run_state_machine_as_test(FailsEventually)
    try:
        FailsEventually.TestCase.settings = Settings(
            FailsEventually.TestCase.settings, stateful_step_count=15
        )
        run_state_machine_as_test(
            FailsEventually, settings=Settings(stateful_step_count=2)
        )
    finally:
        FailsEventually.TestCase.settings = Settings(
            FailsEventually.TestCase.settings, stateful_step_count=5
        )


def test_settings_argument_is_validated():
    with pytest.raises(InvalidArgument):
        run_state_machine_as_test(FailsEventually, settings=object())


def test_runner_that_checks_factory_produced_a_machine():
    with pytest.raises(InvalidArgument):
        run_state_machine_as_test(object)


@skipif_threading
def test_settings_attribute_is_validated():
    real_settings = FailsEventually.TestCase.settings
    try:
        FailsEventually.TestCase.settings = object()
        with pytest.raises(InvalidArgument):
            run_state_machine_as_test(FailsEventually)
    finally:
        FailsEventually.TestCase.settings = real_settings


def test_saves_failing_example_in_database():
    db = InMemoryExampleDatabase()
    ss = Settings(
        database=db, max_examples=1000, suppress_health_check=list(HealthCheck)
    )
    with raises(AssertionError):
        run_state_machine_as_test(DepthMachine, settings=ss)
    assert any(list(db.data.values()))


def test_can_run_with_no_db():
    with raises(AssertionError):
        run_state_machine_as_test(
            DepthMachine, settings=Settings(database=None, max_examples=10_000)
        )


def test_stateful_double_rule_is_forbidden(recwarn):
    with pytest.raises(InvalidDefinition):

        class DoubleRuleMachine(RuleBasedStateMachine):
            @rule(num=just(1))
            @rule(num=just(2))
            def whatevs(self, num):
                pass


def test_can_explicitly_call_functions_when_precondition_not_satisfied():
    class BadPrecondition(RuleBasedStateMachine):
        def __init__(self):
            super().__init__()

        @precondition(lambda self: False)
        @rule()
        def test_blah(self):
            raise ValueError

        @rule()
        def test_foo(self):
            self.test_blah()

    with pytest.raises(ValueError):
        run_state_machine_as_test(BadPrecondition)


def test_invariant():
    """If an invariant raise an exception, the exception is propagated."""

    class Invariant(RuleBasedStateMachine):
        def __init__(self):
            super().__init__()

        @invariant()
        def test_blah(self):
            raise ValueError

        @rule()
        def do_stuff(self):
            pass

    with pytest.raises(ValueError):
        run_state_machine_as_test(Invariant)


def test_no_double_invariant():
    """The invariant decorator can't be applied multiple times to a single
    function."""
    with raises(InvalidDefinition):

        class Invariant(RuleBasedStateMachine):
            def __init__(self):
                super().__init__()

            @invariant()
            @invariant()
            def test_blah(self):
                pass

            @rule()
            def do_stuff(self):
                pass


def test_invariant_precondition():
    """If an invariant precodition isn't met, the invariant isn't run.

    The precondition decorator can be applied in any order.
    """

    class Invariant(RuleBasedStateMachine):
        def __init__(self):
            super().__init__()

        @invariant()
        @precondition(lambda _: False)
        def an_invariant(self):
            raise ValueError

        @precondition(lambda _: False)
        @invariant()
        def another_invariant(self):
            raise ValueError

        @rule()
        def do_stuff(self):
            pass

    run_state_machine_as_test(Invariant)


@pytest.mark.parametrize(
    "decorators",
    [
        (invariant(), rule()),
        (rule(), invariant()),
        (invariant(), initialize()),
        (initialize(), invariant()),
        (invariant(), precondition(lambda self: True), rule()),
        (rule(), precondition(lambda self: True), invariant()),
        (precondition(lambda self: True), invariant(), rule()),
        (precondition(lambda self: True), rule(), invariant()),
    ],
    ids=lambda x: "-".join(f.__qualname__.split(".")[0] for f in x),
)
def test_invariant_and_rule_are_incompatible(decorators):
    """It's an error to apply @invariant and @rule to the same method."""

    def method(self):
        pass

    for d in decorators[:-1]:
        method = d(method)
    with pytest.raises(InvalidDefinition):
        decorators[-1](method)


def test_invalid_rule_argument():
    """Rule kwargs that are not a Strategy are expected to raise an InvalidArgument error."""
    with pytest.raises(InvalidArgument):

        class InvalidRuleMachine(RuleBasedStateMachine):
            @rule(strategy=object())
            def do_stuff(self):
                pass


def test_invalid_initialize_argument():
    """Initialize kwargs that are not a Strategy are expected to raise an InvalidArgument error."""
    with pytest.raises(InvalidArgument):

        class InvalidInitialize(RuleBasedStateMachine):
            @initialize(strategy=object())
            def initialize(self):
                pass


def test_multiple_invariants():
    """If multiple invariants are present, they all get run."""

    class Invariant(RuleBasedStateMachine):
        def __init__(self):
            super().__init__()
            self.first_invariant_ran = False

        @invariant()
        def invariant_1(self):
            self.first_invariant_ran = True

        @precondition(lambda self: self.first_invariant_ran)
        @invariant()
        def invariant_2(self):
            raise ValueError

        @rule()
        def do_stuff(self):
            pass

    with pytest.raises(ValueError):
        run_state_machine_as_test(Invariant)


def test_explicit_invariant_call_with_precondition():
    """Invariants can be called explicitly even if their precondition is not
    satisfied."""

    class BadPrecondition(RuleBasedStateMachine):
        def __init__(self):
            super().__init__()

        @precondition(lambda self: False)
        @invariant()
        def test_blah(self):
            raise ValueError

        @rule()
        def test_foo(self):
            self.test_blah()

    with pytest.raises(ValueError):
        run_state_machine_as_test(BadPrecondition)


def test_invariant_checks_initial_state_if_no_initialize_rules():
    """Invariants are checked before any rules run."""

    class BadPrecondition(RuleBasedStateMachine):
        def __init__(self):
            super().__init__()
            self.num = 0

        @invariant()
        def test_blah(self):
            if self.num == 0:
                raise ValueError

        @rule()
        def test_foo(self):
            self.num += 1

    with pytest.raises(ValueError):
        run_state_machine_as_test(BadPrecondition)


def test_invariant_failling_present_in_falsifying_example():
    @Settings(print_blob=False)
    class BadInvariant(RuleBasedStateMachine):
        @initialize()
        def initialize_1(self):
            pass

        @invariant()
        def invariant_1(self):
            raise ValueError

        @rule()
        def rule_1(self):
            pass

    with pytest.raises(ValueError) as err:
        run_state_machine_as_test(BadInvariant)

    result = "\n".join(err.value.__notes__)
    assert (
        result
        == """
Falsifying example:
state = BadInvariant()
state.initialize_1()
state.invariant_1()
state.teardown()
""".strip()
    )


def test_invariant_present_in_falsifying_example():
    @Settings(print_blob=False, phases=tuple(Phase)[:-1])
    class BadRuleWithGoodInvariants(RuleBasedStateMachine):
        def __init__(self):
            super().__init__()
            self.num = 0

        @initialize()
        def initialize_1(self):
            pass

        @invariant(check_during_init=True)
        def invariant_1(self):
            pass

        @invariant(check_during_init=False)
        def invariant_2(self):
            pass

        @precondition(lambda self: self.num > 0)
        @invariant()
        def invariant_3(self):
            pass

        @rule()
        def rule_1(self):
            self.num += 1
            if self.num == 2:
                raise ValueError

    with pytest.raises(ValueError) as err:
        run_state_machine_as_test(BadRuleWithGoodInvariants)

    expected = """
Falsifying example:
state = BadRuleWithGoodInvariants()
state.invariant_1()
state.initialize_1()
state.invariant_1()
state.invariant_2()
state.rule_1()
state.invariant_1()
state.invariant_2()
state.invariant_3()
state.rule_1()
state.teardown()
""".strip()

    result = "\n".join(err.value.__notes__).strip()
    assert expected == result


def test_always_runs_at_least_one_step():
    class CountSteps(RuleBasedStateMachine):
        def __init__(self):
            super().__init__()
            self.count = 0

        @rule()
        def do_something(self):
            self.count += 1

        def teardown(self):
            assert self.count > 0

    run_state_machine_as_test(CountSteps)


def test_removes_needless_steps():
    """Regression test from an example based on
    tests/nocover/test_database_agreement.py, but without the expensive bits.
    Comparing two database implementations in which deletion is broken, so as
    soon as a key/value pair is successfully deleted the test will now fail if
    you ever check that key.

    The main interesting feature of this is that it has a lot of
    opportunities to generate keys and values before it actually fails,
    but will still fail with very high probability.
    """

    @Settings(derandomize=True, max_examples=1000, deadline=None)
    class IncorrectDeletion(RuleBasedStateMachine):
        def __init__(self):
            super().__init__()
            self.__saved = defaultdict(set)
            self.__deleted = defaultdict(set)

        keys = Bundle("keys")
        values = Bundle("values")

        @rule(target=keys, k=binary())
        def k(self, k):
            return k

        @rule(target=values, v=binary())
        def v(self, v):
            return v

        @rule(k=keys, v=values)
        def save(self, k, v):
            self.__saved[k].add(v)

        @rule(k=keys, v=values)
        def delete(self, k, v):
            if v in self.__saved[k]:
                self.__deleted[k].add(v)

        @rule(k=keys)
        def values_agree(self, k):
            assert not self.__deleted[k]

    with pytest.raises(AssertionError) as err:
        run_state_machine_as_test(IncorrectDeletion)

    result = "\n".join(err.value.__notes__)
    assert result.count(" = state.k(") == 1
    assert result.count(" = state.v(") == 1


def test_prints_equal_values_with_correct_variable_name():
    @Settings(max_examples=100, suppress_health_check=list(HealthCheck))
    class MovesBetweenBundles(RuleBasedStateMachine):
        b1 = Bundle("b1")
        b2 = Bundle("b2")

        @rule(target=b1)
        def create(self):
            return []

        @rule(target=b2, source=b1)
        def transfer(self, source):
            return source

        @rule(source=b2)
        def fail(self, source):
            raise AssertionError

    with pytest.raises(AssertionError) as err:
        run_state_machine_as_test(MovesBetweenBundles)

    result = "\n".join(err.value.__notes__)
    for m in ["create", "transfer", "fail"]:
        assert result.count("state." + m) == 1
    assert "b1_0 = state.create()" in result
    assert "b2_0 = state.transfer(source=b1_0)" in result
    assert "state.fail(source=b2_0)" in result


def test_initialize_rule():
    @Settings(max_examples=1000)
    class WithInitializeRules(RuleBasedStateMachine):
        initialized: ClassVar = []

        @initialize()
        def initialize_a(self):
            self.initialized.append("a")

        @initialize()
        def initialize_b(self):
            self.initialized.append("b")

        @initialize()
        def initialize_c(self):
            self.initialized.append("c")

        @rule()
        def fail_fast(self):
            raise AssertionError

    with pytest.raises(AssertionError) as err:
        run_state_machine_as_test(WithInitializeRules)

    assert set(WithInitializeRules.initialized[-3:]) == {"a", "b", "c"}
    result = err.value.__notes__[1:]
    assert result[0] == "state = WithInitializeRules()"
    # Initialize rules call order is shuffled
    assert {result[1], result[2], result[3]} == {
        "state.initialize_a()",
        "state.initialize_b()",
        "state.initialize_c()",
    }
    assert result[4] == "state.fail_fast()"
    assert result[5] == "state.teardown()"


def test_initialize_rule_populate_bundle():
    class WithInitializeBundleRules(RuleBasedStateMachine):
        a = Bundle("a")

        @initialize(target=a, dep=just("dep"))
        def initialize_a(self, dep):
            return f"a a_0 with ({dep})"

        @rule(param=a)
        def fail_fast(self, param):
            raise AssertionError

    WithInitializeBundleRules.TestCase.settings = NO_BLOB_SETTINGS
    with pytest.raises(AssertionError) as err:
        run_state_machine_as_test(WithInitializeBundleRules)

    result = "\n".join(err.value.__notes__)
    assert (
        result
        == """
Falsifying example:
state = WithInitializeBundleRules()
a_0 = state.initialize_a(dep='dep')
state.fail_fast(param=a_0)
state.teardown()
""".strip()
    )


def test_initialize_rule_dont_mix_with_precondition():
    with pytest.raises(
        InvalidDefinition,
        match=(
            "BadStateMachine\\.initialize has been decorated with both @initialize "
            "and @precondition"
        ),
    ):

        class BadStateMachine(RuleBasedStateMachine):
            @precondition(lambda self: True)
            @initialize()
            def initialize(self):
                pass

    # Also test decorator application in reverse order

    with pytest.raises(
        InvalidDefinition,
        match=(
            "BadStateMachineReverseOrder\\.initialize has been decorated with both "
            "@initialize and @precondition"
        ),
    ):

        class BadStateMachineReverseOrder(RuleBasedStateMachine):
            @initialize()
            @precondition(lambda self: True)
            def initialize(self):
                pass


def test_initialize_rule_dont_mix_with_regular_rule():
    with pytest.raises(
        InvalidDefinition,
        match="BadStateMachine\\.initialize has been decorated with both @rule and @initialize",
    ):

        class BadStateMachine(RuleBasedStateMachine):
            @rule()
            @initialize()
            def initialize(self):
                pass

    with pytest.raises(
        InvalidDefinition,
        match=(
            "BadStateMachineReverseOrder\\.initialize has been decorated with both "
            "@rule and @initialize"
        ),
    ):

        class BadStateMachineReverseOrder(RuleBasedStateMachine):
            @initialize()
            @rule()
            def initialize(self):
                pass


def test_initialize_rule_cannot_be_double_applied():
    with pytest.raises(
        InvalidDefinition,
        match="BadStateMachine\\.initialize has been decorated with @initialize twice",
    ):

        class BadStateMachine(RuleBasedStateMachine):
            @initialize()
            @initialize()
            def initialize(self):
                pass


def test_initialize_rule_in_state_machine_with_inheritance():
    class ParentStateMachine(RuleBasedStateMachine):
        initialized: ClassVar = []

        @initialize()
        def initialize_a(self):
            self.initialized.append("a")

    class ChildStateMachine(ParentStateMachine):
        @initialize()
        def initialize_b(self):
            self.initialized.append("b")

        @rule()
        def fail_fast(self):
            raise AssertionError

    with pytest.raises(AssertionError) as err:
        run_state_machine_as_test(ChildStateMachine)

    assert set(ChildStateMachine.initialized[-2:]) == {"a", "b"}
    result = err.value.__notes__[1:]
    assert result[0] == "state = ChildStateMachine()"
    # Initialize rules call order is shuffled
    assert {result[1], result[2]} == {"state.initialize_a()", "state.initialize_b()"}
    assert result[3] == "state.fail_fast()"
    assert result[4] == "state.teardown()"


def test_can_manually_call_initialize_rule():
    class StateMachine(RuleBasedStateMachine):
        initialize_called_counter = 0

        @initialize()
        def initialize(self):
            self.initialize_called_counter += 1

        @rule()
        def fail_eventually(self):
            self.initialize()
            assert self.initialize_called_counter <= 2

    StateMachine.TestCase.settings = NO_BLOB_SETTINGS
    with pytest.raises(AssertionError) as err:
        run_state_machine_as_test(StateMachine)

    result = "\n".join(err.value.__notes__)
    assert (
        result
        == """
Falsifying example:
state = StateMachine()
state.initialize()
state.fail_eventually()
state.fail_eventually()
state.teardown()
""".strip()
    )


def test_steps_printed_despite_pytest_fail():
    # Test for https://github.com/HypothesisWorks/hypothesis/issues/1372
    @Settings(print_blob=False)
    class RaisesProblem(RuleBasedStateMachine):
        @rule()
        def oops(self):
            pytest.fail("note that this raises a BaseException")

    with pytest.raises(Failed) as err:
        run_state_machine_as_test(RaisesProblem)
    assert (
        "\n".join(err.value.__notes__).strip()
        == """
Falsifying example:
state = RaisesProblem()
state.oops()
state.teardown()""".strip()
    )


def test_steps_not_printed_with_pytest_skip(capsys):
    class RaisesProblem(RuleBasedStateMachine):
        @rule()
        def skip_whole_test(self):
            pytest.skip()

    with pytest.raises(Skipped):
        run_state_machine_as_test(RaisesProblem)
    out, _ = capsys.readouterr()
    assert "state" not in out


def test_rule_deprecation_targets_and_target():
    k, v = Bundle("k"), Bundle("v")
    with pytest.raises(InvalidArgument):
        rule(targets=(k,), target=v)


def test_rule_deprecation_bundle_by_name():
    Bundle("k")
    with pytest.raises(InvalidArgument):
        rule(target="k")


def test_rule_non_bundle_target():
    with pytest.raises(InvalidArgument):
        rule(target=integers())


def test_rule_non_bundle_target_oneof():
    k, v = Bundle("k"), Bundle("v")
    pattern = r".+ `one_of(a, b)` or `a | b` .+"
    with pytest.raises(InvalidArgument, match=pattern):
        rule(target=k | v)


def test_uses_seed(capsys):
    @seed(0)
    class TrivialMachine(RuleBasedStateMachine):
        @rule()
        def oops(self):
            raise AssertionError

    with pytest.raises(AssertionError):
        run_state_machine_as_test(TrivialMachine)
    out, _ = capsys.readouterr()
    assert "@seed" not in out


def test_reproduce_failure_works():
    @reproduce_failure(__version__, encode_failure([False, 0, True]))
    class TrivialMachine(RuleBasedStateMachine):
        @rule()
        def oops(self):
            raise AssertionError

    with pytest.raises(AssertionError):
        run_state_machine_as_test(TrivialMachine, settings=Settings(print_blob=True))


def test_reproduce_failure_fails_if_no_error():
    @reproduce_failure(__version__, encode_failure([False, 0, True]))
    class TrivialMachine(RuleBasedStateMachine):
        @rule()
        def ok(self):
            pass

    with pytest.raises(DidNotReproduce):
        run_state_machine_as_test(TrivialMachine, settings=Settings(print_blob=True))


def test_cannot_have_zero_steps():
    with pytest.raises(InvalidArgument):
        Settings(stateful_step_count=0)


def test_arguments_do_not_use_names_of_return_values():
    # See https://github.com/HypothesisWorks/hypothesis/issues/2341
    class TrickyPrintingMachine(RuleBasedStateMachine):
        data = Bundle("data")

        @initialize(target=data, value=integers())
        def init_data(self, value):
            return value

        @rule(d=data)
        def mostly_fails(self, d):
            assert d == 42

    with pytest.raises(AssertionError) as err:
        run_state_machine_as_test(TrickyPrintingMachine)
    assert "data_0 = state.init_data(value=0)" in err.value.__notes__
    assert "data_0 = state.init_data(value=data_0)" not in err.value.__notes__


class TrickyInitMachine(RuleBasedStateMachine):
    @initialize()
    def init_a(self):
        self.a = 0

    @rule()
    def inc(self):
        self.a += 1

    @invariant()
    def check_a_positive(self):
        # This will fail if run before the init_a method, but without
        # @invariant(check_during_init=True) it will only run afterwards.
        assert self.a >= 0


def test_invariants_are_checked_after_init_steps():
    run_state_machine_as_test(TrickyInitMachine)


def test_invariants_can_be_checked_during_init_steps():
    class UndefinedMachine(TrickyInitMachine):
        @invariant(check_during_init=True)
        def check_a_defined(self):
            # This will fail because `a` is undefined before the init rule.
            self.a

    with pytest.raises(AttributeError):
        run_state_machine_as_test(UndefinedMachine)


def test_check_during_init_must_be_boolean():
    invariant(check_during_init=False)
    invariant(check_during_init=True)
    with pytest.raises(InvalidArgument):
        invariant(check_during_init="not a bool")


def test_deprecated_target_consumes_bundle():
    # It would be nicer to raise this error at runtime, but the internals make
    # this sadly impractical.  Most InvalidDefinition errors happen at, well,
    # definition-time already anyway, so it's not *worse* than the status quo.
    with validate_deprecation():
        rule(target=consumes(Bundle("b")))


@Settings(stateful_step_count=5)
class MinStepsMachine(RuleBasedStateMachine):
    @initialize()
    def init_a(self):
        self.a = 0

    @rule()
    def inc(self):
        self.a += 1

    @invariant()
    def not_too_many_steps(self):
        assert self.a < 10

    def teardown(self):
        assert self.a >= 2


# Replay overruns after we trigger a crosshair.util.IgnoreAttempt exception for n=3
@xfail_on_crosshair(Why.other)
def test_min_steps_argument():
    # You must pass a non-negative integer...
    for n_steps in (-1, "nan", 5.0):
        with pytest.raises(InvalidArgument):
            run_state_machine_as_test(MinStepsMachine, _min_steps=n_steps)

    # and if you do, we'll take at least that many steps
    run_state_machine_as_test(MinStepsMachine, _min_steps=3)

    # (oh, and it's OK if you ask for more than we're actually going to take)
    run_state_machine_as_test(MinStepsMachine, _min_steps=20)


class ErrorsOnClassAttributeSettings(RuleBasedStateMachine):
    settings = Settings(derandomize=True)

    @rule()
    def step(self):
        pass


def test_fails_on_settings_class_attribute():
    with pytest.raises(
        InvalidDefinition,
        match=r"Assigning .+ as a class attribute does nothing",
    ):
        run_state_machine_as_test(ErrorsOnClassAttributeSettings)


def test_single_target_multiple():
    class Machine(RuleBasedStateMachine):
        a = Bundle("a")

        @initialize(target=a)
        def initialize(self):
            return multiple("ret1", "ret2", "ret3")

        @rule(param=a)
        def fail_fast(self, param):
            raise AssertionError

    Machine.TestCase.settings = NO_BLOB_SETTINGS
    with pytest.raises(AssertionError) as err:
        run_state_machine_as_test(Machine)

    result = "\n".join(err.value.__notes__)
    assert (
        result
        == """
Falsifying example:
state = Machine()
a_0, a_1, a_2 = state.initialize()
state.fail_fast(param=a_2)
state.teardown()
""".strip()
    )


@pytest.mark.parametrize(
    "bundle_names,initial,repr_",
    [
        ("a", "ret1", "a_0 = state.init()"),
        ("aba", "ret1", "a_0 = b_0 = a_1 = state.init()"),
        ("a", multiple(), "state.init()"),
        ("aba", multiple(), "state.init()"),
        ("a", multiple("ret1"), "(a_0,) = state.init()"),
        ("aba", multiple("ret1"), "(a_0,) = (b_0,) = (a_1,) = state.init()"),
        ("a", multiple("ret1", "ret2"), "a_0, a_1 = state.init()"),
        (
            "aba",
            multiple("ret1", "ret2"),
            "\n".join(  # noqa: FLY002  # no, f-string is not more readable
                [
                    "a_0, a_1 = state.init()",
                    "b_0, b_1 = a_0, a_1",
                    "a_2, a_3 = a_0, a_1",
                ]
            ),
        ),
    ],
)
def test_targets_repr(bundle_names, initial, repr_):
    bundles = {name: Bundle(name) for name in bundle_names}

    class Machine(RuleBasedStateMachine):
        @initialize(targets=[bundles[name] for name in bundle_names])
        def init(self):
            return initial

        @rule()
        def fail_fast(self):
            raise AssertionError

    Machine.TestCase.settings = NO_BLOB_SETTINGS
    with pytest.raises(AssertionError) as err:
        run_state_machine_as_test(Machine)

    result = "\n".join(err.value.__notes__)
    assert (
        result
        == f"""
Falsifying example:
state = Machine()
{repr_}
state.fail_fast()
state.teardown()
""".strip()
    )


def test_multiple_targets():
    class Machine(RuleBasedStateMachine):
        a = Bundle("a")
        b = Bundle("b")

        @initialize(targets=(a, b))
        def initialize(self):
            return multiple("ret1", "ret2", "ret3")

        @rule(
            a1=consumes(a),
            a2=consumes(a),
            a3=consumes(a),
            b1=consumes(b),
            b2=consumes(b),
            b3=consumes(b),
        )
        def fail_fast(self, a1, a2, a3, b1, b2, b3):
            raise AssertionError

    Machine.TestCase.settings = NO_BLOB_SETTINGS
    with pytest.raises(AssertionError) as err:
        run_state_machine_as_test(Machine)

    result = "\n".join(err.value.__notes__)
    assert (
        result
        == """
Falsifying example:
state = Machine()
a_0, a_1, a_2 = state.initialize()
b_0, b_1, b_2 = a_0, a_1, a_2
state.fail_fast(a1=a_2, a2=a_1, a3=a_0, b1=b_2, b2=b_1, b3=b_0)
state.teardown()
""".strip()
    )


def test_multiple_common_targets():
    class Machine(RuleBasedStateMachine):
        a = Bundle("a")
        b = Bundle("b")

        @initialize(targets=(a, b, a))
        def initialize(self):
            return multiple("ret1", "ret2", "ret3")

        @rule(
            a1=consumes(a),
            a2=consumes(a),
            a3=consumes(a),
            a4=consumes(a),
            a5=consumes(a),
            a6=consumes(a),
            b1=consumes(b),
            b2=consumes(b),
            b3=consumes(b),
        )
        def fail_fast(self, a1, a2, a3, a4, a5, a6, b1, b2, b3):
            raise AssertionError

    Machine.TestCase.settings = NO_BLOB_SETTINGS
    with pytest.raises(AssertionError) as err:
        run_state_machine_as_test(Machine)

    result = "\n".join(err.value.__notes__)
    assert (
        result
        == """
Falsifying example:
state = Machine()
a_0, a_1, a_2 = state.initialize()
b_0, b_1, b_2 = a_0, a_1, a_2
a_3, a_4, a_5 = a_0, a_1, a_2
state.fail_fast(a1=a_5, a2=a_4, a3=a_3, a4=a_2, a5=a_1, a6=a_0, b1=b_2, b2=b_1, b3=b_0)
state.teardown()
""".strip()
    )


class LotsOfEntropyPerStepMachine(RuleBasedStateMachine):
    # Regression tests for https://github.com/HypothesisWorks/hypothesis/issues/3618
    @rule(data=binary(min_size=512, max_size=512))
    def rule1(self, data):
        assert data


@pytest.mark.skipif(
    Settings.get_current_profile_name() == "crosshair",
    reason="takes hours; too much symbolic data",
)
def test_lots_of_entropy():
    run_state_machine_as_test(LotsOfEntropyPerStepMachine)


def test_flatmap():
    class Machine(RuleBasedStateMachine):
        buns = Bundle("buns")

        @initialize(target=buns)
        def create_bun(self):
            return 0

        @rule(target=buns, bun=buns.flatmap(lambda x: just(x + 1)))
        def use_flatmap(self, bun):
            assert isinstance(bun, int)
            return bun

        @rule(bun=buns)
        def use_directly(self, bun):
            assert isinstance(bun, int)

    Machine.TestCase.settings = Settings(stateful_step_count=5, max_examples=10)
    run_state_machine_as_test(Machine)


def test_use_bundle_within_other_strategies():
    class Class:
        def __init__(self, value):
            self.value = value

    class Machine(RuleBasedStateMachine):
        my_bundle = Bundle("my_bundle")

        @initialize(target=my_bundle)
        def set_initial(self, /) -> str:
            return "sample text"

        @rule(instance=st.builds(Class, my_bundle))
        def check(self, instance):
            assert isinstance(instance, Class)
            assert isinstance(instance.value, str)

    Machine.TestCase.settings = Settings(stateful_step_count=5, max_examples=10)
    run_state_machine_as_test(Machine)


def test_precondition_cannot_be_used_without_rule():
    class BadStateMachine(RuleBasedStateMachine):
        @precondition(lambda: True)
        def has_precondition_but_no_rule(self):
            pass

        @rule(n=st.integers())
        def trivial(self, n):
            pass

    with pytest.raises(
        InvalidDefinition,
        match=(
            "BadStateMachine\\.has_precondition_but_no_rule has been decorated "
            "with @precondition, but not @rule"
        ),
    ):
        BadStateMachine.TestCase().runTest()



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import pytest

from hypothesistooling.releasemanagement import (
    bump_version_info,
    parse_release_file_contents,
    release_date_string,
    replace_assignment_in_string as replace,
    update_markdown_changelog,
)


def parse_release(contents):
    return parse_release_file_contents(contents, "<string>")


def test_update_single_line():
    assert replace("a = 1", "a", "2") == "a = 2"


def test_update_without_spaces():
    assert replace("a=1", "a", "2") == "a=2"


def test_update_in_middle():
    assert replace("a = 1\nb=2\nc = 3", "b", "4") == "a = 1\nb=4\nc = 3"


def test_quotes_string_to_assign():
    assert replace("a.c = 1", "a.c", "2") == "a.c = 2"
    with pytest.raises(ValueError):
        replace("abc = 1", "a.c", "2")


def test_duplicates_are_errors():
    with pytest.raises(ValueError):
        replace("a = 1\na=1", "a", "2")


def test_missing_is_error():
    with pytest.raises(ValueError):
        replace("", "a", "1")


def test_bump_minor_version():
    assert bump_version_info((1, 1, 1), "minor")[0] == "1.2.0"


def test_parse_release_file():
    assert parse_release("RELEASE_TYPE: patch\nhi") == ("patch", "hi")
    assert parse_release("RELEASE_TYPE: minor\n\n\n\nhi") == ("minor", "hi")
    assert parse_release("RELEASE_TYPE: major\n \n\nhi") == ("major", "hi")


def test_invalid_release():
    with pytest.raises(ValueError):
        parse_release("RELEASE_TYPE: wrong\nstuff")

    with pytest.raises(ValueError):
        parse_release("")


TEST_CHANGELOG = f"""
# A test project 1.2.3 ({release_date_string()})

some stuff happened

# some previous log entry
"""


def test_update_changelog(tmp_path):
    path = tmp_path / "CHANGELOG.md"
    path.write_text("# some previous log entry\n", encoding="utf-8")
    update_markdown_changelog(
        str(path), "A test project", "1.2.3", "some stuff happened"
    )
    assert path.read_text(encoding="utf-8").strip() == TEST_CHANGELOG.strip()


def test_changelog_parsing_strips_trailing_whitespace():
    header = "RELEASE_TYPE: patch\n\n"
    contents = "Adds a feature\n    indented.\n"
    _level, out = parse_release(header + contents.replace("feature", "feature    "))
    assert contents.strip() == out



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from random import Random

from hypothesis import Phase, find, settings, strategies as st

from tests.common.utils import Why, xfail_on_crosshair


@xfail_on_crosshair(Why.symbolic_outside_context)
def test_find_uses_provided_random():
    prev = None

    for _ in range(3):
        seen = None

        def test(v):
            if len(v) > 5:
                nonlocal seen
                if seen is not None:
                    return v == seen
                else:
                    seen = v
                    return True

        result = find(
            st.text(),
            test,
            random=Random(13),
            settings=settings(phases=[Phase.generate], max_examples=1000),
        )
        if prev is None:
            prev = result
        else:
            assert prev == result



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import asyncio
import sys
from collections.abc import Callable

import pytest

from hypothesis import errors, given, reject, strategies as st
from hypothesis.internal.compat import ExceptionGroup
from hypothesis.strategies import DataObject

# this file is not typechecked by mypy, which only runs py310

if sys.version_info < (3, 11):
    pytest.skip("asyncio.TaskGroup not available on <py3.11", allow_module_level=True)


def test_exceptiongroup_discard_frozen():
    """Basic test that raises Frozen+Unsatisfiable.
    Frozen is thrown out, and Unsatisfiable is raised"""

    @given(st.data())
    def test_function(data):
        async def task(pred):
            return data.draw(st.booleans().filter(pred))

        async def _main():
            async with asyncio.TaskGroup() as tg:
                tg.create_task(task(bool))
                tg.create_task(task(lambda _: False))

        asyncio.run(_main())

    with pytest.raises(errors.Unsatisfiable):
        test_function()


def test_exceptiongroup_nested() -> None:
    @given(st.data())
    def test_function(data: DataObject) -> None:
        async def task(pred: Callable[[bool], bool]) -> None:
            data.draw(st.booleans().filter(pred))

        async def _main() -> None:
            async with asyncio.TaskGroup(), asyncio.TaskGroup() as tg2:
                tg2.create_task(task(bool))
                tg2.create_task(task(lambda _: False))

        asyncio.run(_main())

    with pytest.raises(errors.Unsatisfiable):
        test_function()


def test_exceptiongroup_user_originated() -> None:
    @given(st.data())
    def test_function(data):
        raise ExceptionGroup("foo", [ValueError(), ValueError()])

    with pytest.raises(ExceptionGroup) as exc_info:
        test_function()
    e = exc_info.value
    assert e.message == "foo"
    assert isinstance(e, ExceptionGroup)
    assert len(e.exceptions) == 2
    assert all(isinstance(child_e, ValueError) for child_e in e.exceptions)

    @given(st.data())
    def test_single_exc_group(data):
        raise ExceptionGroup("important message for user", [ValueError()])

    with pytest.raises(ExceptionGroup) as exc_info:
        test_single_exc_group()
    e = exc_info.value
    assert e.message == "important message for user"
    assert isinstance(e, ExceptionGroup)
    assert len(e.exceptions) == 1
    assert isinstance(e.exceptions[0], ValueError)


def test_exceptiongroup_multiple_stop() -> None:
    @given(st.data())
    def test_function(data):
        async def task(d: DataObject) -> None:
            d.conjecture_data.mark_invalid()

        async def _main(d: DataObject) -> None:
            async with asyncio.TaskGroup() as tg:
                tg.create_task(task(d))
                tg.create_task(task(d))

        asyncio.run(_main(data))

    # multiple stoptests -> single stoptest -> unsatisfiable
    with pytest.raises(errors.Unsatisfiable):
        test_function()


def test_exceptiongroup_stop_and_hypothesisexception() -> None:
    # group with stoptest+invalidargument -> invalidargument
    @given(st.data())
    def test_function(data):
        async def task_stoptest(d: DataObject) -> None:
            # only mark some runs as invalid to not raise Unsatisfiable
            if d.draw(st.integers(min_value=0, max_value=1)) == 1:
                d.conjecture_data.mark_invalid()

        async def task_invalid_argument(d: DataObject) -> None:
            d.draw(st.integers(max_value=2, min_value=3))

        async def _main(d: DataObject) -> None:
            async with asyncio.TaskGroup() as tg:
                tg.create_task(task_stoptest(d))
                tg.create_task(task_invalid_argument(d))

        asyncio.run(_main(data))

    with pytest.raises(errors.InvalidArgument):
        test_function()


def test_exceptiongroup_multiple_hypothesisexception() -> None:
    # multiple UnsatisfiedAssumption => first one is reraised => engine suppresses it

    @given(st.integers(min_value=0, max_value=3))
    def test_function(val: int) -> None:
        async def task(value: int) -> None:
            if value == 0:
                reject()

        async def _main(value: int) -> None:
            async with asyncio.TaskGroup() as tg:
                tg.create_task(task(value))
                tg.create_task(task(value))

        asyncio.run(_main(val))

    test_function()


def test_exceptiongroup_multiple_InvalidArgument() -> None:
    # multiple InvalidArgument => only first one is reraised... which seems bad.
    # But raising a group might break ghostwriter(?)

    @given(st.data())
    def test_function(data: DataObject) -> None:
        async def task1(d: DataObject) -> None:
            d.draw(st.integers(max_value=1, min_value=3))

        async def task2(d: DataObject) -> None:
            d.draw(st.integers(max_value=2, min_value=3))

        async def _main(d: DataObject) -> None:
            async with asyncio.TaskGroup() as tg:
                tg.create_task(task1(d))
                tg.create_task(task2(d))

        asyncio.run(_main(data))

    with pytest.raises(errors.InvalidArgument):
        test_function()



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import sys

import pytest

from hypothesis import HealthCheck, Verbosity, assume, example, given, reject, settings
from hypothesis.core import StateForActualGivenExecution
from hypothesis.errors import Flaky, FlakyFailure, Unsatisfiable, UnsatisfiedAssumption
from hypothesis.internal.compat import ExceptionGroup
from hypothesis.internal.conjecture.engine import MIN_TEST_CALLS
from hypothesis.internal.scrutineer import Tracer
from hypothesis.strategies import booleans, composite, integers, lists, random_module

from tests.common.utils import Why, no_shrink, skipif_threading, xfail_on_crosshair


class Nope(Exception):
    pass


def test_fails_only_once_is_flaky():
    first_call = True

    @given(integers())
    def rude(x):
        nonlocal first_call
        if first_call:
            first_call = False
            raise Nope

    with pytest.raises(FlakyFailure, match="Falsified on the first call but") as e:
        rude()
    exceptions = e.value.exceptions
    assert len(exceptions) == 1
    assert isinstance(exceptions[0], Nope)


def test_fails_differently_is_flaky():
    call_count = 0

    class DifferentNope(Exception):
        pass

    @given(integers())
    @settings(database=None)
    def rude(x):
        nonlocal call_count
        if x == 0:
            call_count += 1
            if call_count > 1:
                raise Nope
            else:
                raise DifferentNope

    with pytest.raises(FlakyFailure, match="Inconsistent results from replaying") as e:
        rude()
    exceptions = e.value.exceptions
    assert len(exceptions) == 2
    assert set(map(type, exceptions)) == {Nope, DifferentNope}


@skipif_threading  # executing into global scope
@pytest.mark.skipif(sys.version_info < (3, 11), reason="except* syntax")
def test_exceptiongroup_wrapped_naked_exception_is_flaky():

    # Defer parsing until runtime, as "except*" is syntax error pre 3.11
    rude_def = """
first_call = True
def rude_fn(x):
    global first_call
    if first_call:
        first_call = False
        try:
            raise Nope
        except* Nope:
            raise
    """
    exec(rude_def, globals())
    rude = given(integers())(rude_fn)  # noqa: F821 # defined by exec()

    with pytest.raises(FlakyFailure, match="Falsified on the first call but") as e:
        rude()
    exceptions = e.value.exceptions
    assert list(map(type, exceptions)) == [ExceptionGroup]
    assert list(map(type, exceptions[0].exceptions)) == [Nope]


def test_gives_flaky_error_if_assumption_is_flaky():
    seen = set()

    @given(integers())
    @settings(verbosity=Verbosity.quiet, database=None)
    def oops(s):
        assume(s not in seen)
        seen.add(s)
        raise AssertionError

    with pytest.raises(FlakyFailure, match="Inconsistent results from replaying") as e:
        oops()
    exceptions = e.value.exceptions
    assert len(exceptions) == 2
    assert isinstance(exceptions[0], AssertionError)
    assert isinstance(exceptions[1], UnsatisfiedAssumption)


def test_flaky_with_context_when_fails_only_under_tracing(monkeypatch):
    # make anything fail under tracing
    monkeypatch.setattr(Tracer, "can_trace", staticmethod(lambda: True))
    monkeypatch.setattr(Tracer, "__enter__", lambda *_: 1 / 0)
    # ensure tracing is always entered inside _execute_once_for_engine
    monkeypatch.setattr(StateForActualGivenExecution, "_should_trace", lambda _: True)

    @given(integers())
    def test(x):
        pass

    with pytest.raises(
        FlakyFailure, match="failed on the first run but now succeeds"
    ) as e:
        test()
    exceptions = e.value.exceptions
    assert len(exceptions) == 1
    assert isinstance(exceptions[0], ZeroDivisionError)


@xfail_on_crosshair(Why.symbolic_outside_context)
def test_does_not_attempt_to_shrink_flaky_errors():
    values = []

    @settings(database=None)
    @given(integers())
    def test(x):
        values.append(x)
        assert len(values) != 1

    with pytest.raises(FlakyFailure):
        test()
    # We try a total of ten calls in the generation phase, each usually a
    # unique value, looking briefly (and unsuccessfully) for another bug.
    assert 1 < len(set(values)) <= MIN_TEST_CALLS
    # We don't try any new values while shrinking, just execute the test
    # twice more (to check for flakiness and to raise the bug to the user).
    assert set(values) == set(values[:-2])


class SatisfyMe(Exception):
    pass


@composite
def single_bool_lists(draw):
    n = draw(integers(0, 20))
    result = [False] * (n + 1)
    result[n] = True
    return result


@xfail_on_crosshair(Why.nested_given)
@example([True, False, False, False], [3], None)
@example([False, True, False, False], [3], None)
@example([False, False, True, False], [3], None)
@example([False, False, False, True], [3], None)
@settings(
    deadline=None, suppress_health_check=[HealthCheck.nested_given], max_examples=10
)
@given(lists(booleans()) | single_bool_lists(), lists(integers(1, 3)), random_module())
def test_failure_sequence_inducing(building, testing, rnd):
    buildit = iter(building)
    testit = iter(testing)

    def build(x):
        try:
            assume(not next(buildit))
        except StopIteration:
            pass
        return x

    @given(integers().map(build))
    @settings(
        verbosity=Verbosity.quiet,
        database=None,
        suppress_health_check=list(HealthCheck),
        phases=no_shrink,
        max_examples=10,
    )
    def test(x):
        try:
            i = next(testit)
        except StopIteration:
            return
        if i == 1:
            return
        elif i == 2:
            reject()
        else:
            raise Nope

    try:
        test()
    except (Nope, Flaky, Unsatisfiable):
        pass
    except UnsatisfiedAssumption:
        raise SatisfyMe from None



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import dataclasses
import sys
import typing

import pytest

from hypothesis import given, strategies as st
from hypothesis.errors import InvalidArgument

from tests.common.debug import (
    assert_all_examples,
    assert_simple_property,
    check_can_generate_examples,
    find_any,
)
from tests.common.utils import temp_registered

# Union[A, B] is not equivalent to A | B until 3.14. We'll continue to test both
# until then.
# ruff: noqa: UP007


@pytest.mark.parametrize(
    "annotated_type,expected_strategy_repr",
    [
        (typing.Annotated[int, "foo"], "integers()"),
        (typing.Annotated[list[float], "foo"], "lists(floats())"),
        (typing.Annotated[typing.Annotated[str, "foo"], "bar"], "text()"),
        (
            typing.Annotated[typing.Annotated[list[dict[str, bool]], "foo"], "bar"],
            "lists(dictionaries(keys=text(), values=booleans()))",
        ),
    ],
)
def test_typing_Annotated(annotated_type, expected_strategy_repr):
    assert repr(st.from_type(annotated_type)) == expected_strategy_repr


PositiveInt = typing.Annotated[int, st.integers(min_value=1)]
MoreThanTenInt = typing.Annotated[PositiveInt, st.integers(min_value=10 + 1)]
WithTwoStrategies = typing.Annotated[int, st.integers(), st.none()]
ExtraAnnotationNoStrategy = typing.Annotated[PositiveInt, "metadata"]


def arg_positive(x: PositiveInt):
    assert x > 0


def arg_more_than_ten(x: MoreThanTenInt):
    assert x > 10


@given(st.data())
def test_annotated_positive_int(data):
    data.draw(st.builds(arg_positive))


@given(st.data())
def test_annotated_more_than_ten(data):
    data.draw(st.builds(arg_more_than_ten))


@given(st.data())
def test_annotated_with_two_strategies(data):
    assert data.draw(st.from_type(WithTwoStrategies)) is None


@given(st.data())
def test_annotated_extra_metadata(data):
    assert data.draw(st.from_type(ExtraAnnotationNoStrategy)) > 0


@dataclasses.dataclass
class User:
    id: int
    following: list["User"]  # works with typing.List


@pytest.mark.skipif(sys.version_info[:2] >= (3, 11), reason="works in new Pythons")
def test_string_forward_ref_message():
    # See https://github.com/HypothesisWorks/hypothesis/issues/3016
    s = st.builds(User)
    with pytest.raises(InvalidArgument, match="`from __future__ import annotations`"):
        check_can_generate_examples(s)


@pytest.mark.parametrize("typ", (typing.Union[list[int], int], list[int] | int))
def test_issue_3080(typ):
    # Check for https://github.com/HypothesisWorks/hypothesis/issues/3080
    s = st.from_type(typ)
    find_any(s, lambda x: isinstance(x, int))
    find_any(s, lambda x: isinstance(x, list))


@dataclasses.dataclass
class TypingTuple:
    a: dict[tuple[int, int], str]


@dataclasses.dataclass
class BuiltinTuple:
    a: dict[tuple[int, int], str]


TestDataClass = typing.Union[TypingTuple, BuiltinTuple]


@pytest.mark.parametrize("data_class", [TypingTuple, BuiltinTuple])
@given(data=st.data())
def test_from_type_with_tuple_works(data, data_class: TestDataClass):
    value: TestDataClass = data.draw(st.from_type(data_class))
    assert len(value.a) >= 0


def _shorter_lists(list_type):
    return st.lists(st.from_type(*typing.get_args(list_type)), max_size=2)


def test_can_register_builtin_list():
    # Regression test for https://github.com/HypothesisWorks/hypothesis/issues/3635
    with temp_registered(list, _shorter_lists):
        assert_all_examples(
            st.from_type(list[int]),
            lambda ls: len(ls) <= 2 and {type(x) for x in ls}.issubset({int}),
        )


T = typing.TypeVar("T")


@typing.runtime_checkable
class Fooable(typing.Protocol[T]):
    def foo(self): ...


class FooableConcrete(tuple):
    def foo(self):
        pass


def test_only_tuple_subclasses_in_typing_type():
    # A generic typing type (such as Fooable) whose only concrete
    # instantiations are tuples should still generate tuples. This is in
    # contrast to test_tuple_subclasses_not_generic_sequences, which discards
    # tuples if there are any alternatives.
    with temp_registered(FooableConcrete, st.builds(FooableConcrete)):
        s = st.from_type(Fooable[int])
        assert_all_examples(s, lambda x: type(x) is FooableConcrete)


def test_lookup_registered_tuple():
    sentinel = object()
    typ = tuple[int]
    with temp_registered(tuple, st.just(sentinel)):
        assert_simple_property(st.from_type(typ), lambda v: v is sentinel)
    assert_simple_property(st.from_type(typ), lambda v: v is not sentinel)


sentinel = object()


class LazyStrategyAnnotation:
    __is_annotated_types_grouped_metadata__ = True

    def __iter__(self):
        return iter([st.just(sentinel)])


@given(...)
def test_grouped_protocol_strategy(x: typing.Annotated[int, LazyStrategyAnnotation()]):
    assert x is sentinel


def test_collections_abc_callable_none():
    # https://github.com/HypothesisWorks/hypothesis/issues/4192
    s = st.from_type(collections.abc.Callable[[None], None])
    assert_all_examples(s, lambda x: callable(x) and x(None) is None)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import re

import pytest

pytest_plugins = "pytester"


CONFTEST = """
import time
import hypothesis.internal.conjecture.providers as providers

_called = False

def slow_get_local_constants():
    global _called
    if not _called:
        _called = True
        time.sleep(0.1)
    return providers._local_constants

providers._get_local_constants = slow_get_local_constants
providers._sys_modules_len = None
"""


TESTSUITE = """
from hypothesis import given, Phase, settings
from hypothesis.strategies import integers
from hypothesis.internal.conjecture.providers import _get_local_constants

def test_first():
    # Force constant collection to happen during this test
    _get_local_constants()

@given(integers())
@settings(phases=[Phase.generate], max_examples=1, database=None)
def test_second(x):
    pass
"""


@pytest.mark.parametrize("plugin_disabled", [False, True])
def test_constant_collection_timing(testdir, plugin_disabled):
    # See https://github.com/HypothesisWorks/hypothesis/issues/4627
    testdir.makeconftest(CONFTEST)
    testdir.makepyfile(TESTSUITE)

    args = ["--durations=0", "-vv"]
    if plugin_disabled:
        args += ["-p", "no:hypothesispytest"]

    result = testdir.runpytest(*args)
    result.assert_outcomes(passed=2)

    output = "\n".join(result.stdout.lines)
    match = re.search(r"([\d.]+)s call\s+\S+::test_first", output)
    assert match, f"Could not find test_first timing in:\n{output}"
    test_first_time = float(match.group(1))

    if plugin_disabled:
        assert test_first_time >= 0.05, f"took {test_first_time}s, expected >= 0.05s"
    else:
        assert test_first_time < 0.05, f"took {test_first_time}s, expected < 0.05s"



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import sys
from copy import deepcopy
from datetime import time
from functools import partial, wraps
from inspect import Parameter, Signature, signature
from textwrap import dedent
from unittest.mock import MagicMock, Mock, NonCallableMagicMock, NonCallableMock

import pytest
from pytest import raises

from hypothesis import assume, example, given, strategies as st
from hypothesis.errors import HypothesisWarning
from hypothesis.internal import lambda_sources, reflection
from hypothesis.internal.reflection import (
    convert_keyword_arguments,
    convert_positional_arguments,
    define_function_signature,
    function_digest,
    get_pretty_function_description,
    get_signature,
    is_first_param_referenced_in_function,
    is_identity_function,
    is_mock,
    proxies,
    repr_call,
    required_args,
    source_exec_as_module,
)
from hypothesis.strategies._internal.lazy import LazyStrategy


def do_conversion_test(f, args, kwargs):
    result = f(*args, **kwargs)

    cargs, ckwargs = convert_keyword_arguments(f, args, kwargs)
    assert result == f(*cargs, **ckwargs)

    cargs2, ckwargs2 = convert_positional_arguments(f, args, kwargs)
    assert result == f(*cargs2, **ckwargs2)


def test_simple_conversion():
    def foo(a, b, c):
        return (a, b, c)

    assert convert_keyword_arguments(foo, (1, 2, 3), {}) == ((1, 2, 3), {})
    assert convert_keyword_arguments(foo, (), {"a": 3, "b": 2, "c": 1}) == (
        (3, 2, 1),
        {},
    )

    do_conversion_test(foo, (1, 0), {"c": 2})
    do_conversion_test(foo, (1,), {"c": 2, "b": "foo"})


def test_leaves_unknown_kwargs_in_dict():
    def bar(x, **kwargs):
        pass

    assert convert_keyword_arguments(bar, (1,), {"foo": "hi"}) == ((1,), {"foo": "hi"})
    assert convert_keyword_arguments(bar, (), {"x": 1, "foo": "hi"}) == (
        (1,),
        {"foo": "hi"},
    )
    do_conversion_test(bar, (1,), {})
    do_conversion_test(bar, (), {"x": 1, "y": 1})


def test_errors_on_bad_kwargs():
    def bar():
        pass

    with raises(TypeError):
        convert_keyword_arguments(bar, (), {"foo": 1})


def test_passes_varargs_correctly():
    def foo(*args):
        pass

    assert convert_keyword_arguments(foo, (1, 2, 3), {}) == ((1, 2, 3), {})

    do_conversion_test(foo, (1, 2, 3), {})


def test_errors_if_keyword_precedes_positional():
    def foo(x, y):
        pass

    with raises(TypeError):
        convert_keyword_arguments(foo, (1,), {"x": 2})


def test_errors_if_not_enough_args():
    def foo(a, b, c, d=1):
        pass

    with raises(TypeError):
        convert_keyword_arguments(foo, (1, 2), {"d": 4})


def test_errors_on_extra_kwargs():
    def foo(a):
        pass

    with raises(TypeError, match="keyword"):
        convert_keyword_arguments(foo, (1,), {"b": 1})

    with raises(TypeError, match="keyword"):
        convert_keyword_arguments(foo, (1,), {"b": 1, "c": 2})


def test_positional_errors_if_too_many_args():
    def foo(a):
        pass

    with raises(TypeError, match="too many positional arguments"):
        convert_positional_arguments(foo, (1, 2), {})


def test_positional_errors_if_too_few_args():
    def foo(a, b, c):
        pass

    with raises(TypeError):
        convert_positional_arguments(foo, (1, 2), {})


def test_positional_does_not_error_if_extra_args_are_kwargs():
    def foo(a, b, c):
        pass

    convert_positional_arguments(foo, (1, 2), {"c": 3})


def test_positional_errors_if_given_bad_kwargs():
    def foo(a):
        pass

    with raises(TypeError, match="missing a required argument: 'a'"):
        convert_positional_arguments(foo, (), {"b": 1})


def test_positional_errors_if_given_duplicate_kwargs():
    def foo(a):
        pass

    with raises(TypeError, match="multiple values"):
        convert_positional_arguments(foo, (2,), {"a": 1})


def test_names_of_functions_are_pretty():
    assert (
        get_pretty_function_description(test_names_of_functions_are_pretty)
        == "test_names_of_functions_are_pretty"
    )


class Foo:
    @classmethod
    def bar(cls):
        pass

    def baz(cls):
        pass

    def __repr__(self):
        return "SoNotFoo()"


def test_class_names_are_not_included_in_class_method_prettiness():
    assert get_pretty_function_description(Foo.bar) == "bar"


def test_repr_is_included_in_bound_method_prettiness():
    assert get_pretty_function_description(Foo().baz) == "SoNotFoo().baz"


def test_class_is_not_included_in_unbound_method():
    assert get_pretty_function_description(Foo.baz) == "baz"


def test_does_not_error_on_confused_sources():
    def ed(f, *args):
        return f

    x = ed(
        lambda x, y: (x * y).conjugate() == x.conjugate() * y.conjugate(),
        complex,
        complex,
    )

    get_pretty_function_description(x)


def test_digests_are_reasonably_unique():
    assert function_digest(test_simple_conversion) != function_digest(
        test_does_not_error_on_confused_sources
    )


def test_digest_returns_the_same_value_for_two_calls():
    assert function_digest(test_simple_conversion) == function_digest(
        test_simple_conversion
    )


def test_can_digest_a_built_in_function():
    import math

    assert function_digest(math.isnan) != function_digest(range)


def test_can_digest_a_unicode_lambda():
    function_digest(lambda x: "☃" in str(x))


def test_can_digest_a_function_with_no_name():
    def foo(x, y):
        pass

    function_digest(partial(foo, 1))


def test_arg_string_is_in_order():
    def foo(c, a, b, f, a1):
        pass

    assert repr_call(foo, (1, 2, 3, 4, 5), {}) == "foo(c=1, a=2, b=3, f=4, a1=5)"
    assert (
        repr_call(foo, (1, 2), {"b": 3, "f": 4, "a1": 5})
        == "foo(c=1, a=2, b=3, f=4, a1=5)"
    )


def test_varkwargs_are_sorted_and_after_real_kwargs():
    def foo(d, e, f, **kwargs):
        pass

    assert (
        repr_call(foo, (), {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5, "f": 6})
        == "foo(d=4, e=5, f=6, a=1, b=2, c=3)"
    )


def test_varargs_come_without_equals():
    def foo(a, *args):
        pass

    assert repr_call(foo, (1, 2, 3, 4), {}) == "foo(2, 3, 4, a=1)"


def test_can_mix_varargs_and_varkwargs():
    def foo(*args, **kwargs):
        pass

    assert repr_call(foo, (1, 2, 3), {"c": 7}) == "foo(1, 2, 3, c=7)"


def test_arg_string_does_not_include_unprovided_defaults():
    def foo(a, b, c=9, d=10):
        pass

    assert repr_call(foo, (1,), {"b": 1, "d": 11}) == "foo(a=1, b=1, d=11)"


def universal_acceptor(*args, **kwargs):
    return args, kwargs


def has_one_arg(hello):
    pass


def has_two_args(hello, world):
    pass


def has_a_default(x, y, z=1):
    pass


def has_varargs(*args):
    pass


def has_kwargs(**kwargs):
    pass


@pytest.mark.parametrize("f", [has_one_arg, has_two_args, has_varargs, has_kwargs])
def test_copying_preserves_signature(f):
    af = get_signature(f)
    t = define_function_signature("foo", "docstring", af)(universal_acceptor)
    at = get_signature(t)
    assert af == at


def test_name_does_not_clash_with_function_names():
    def f():
        pass

    @define_function_signature("f", "A docstring for f", signature(f))
    def g():
        pass

    g()


def test_copying_sets_name():
    f = define_function_signature(
        "hello_world", "A docstring for hello_world", signature(has_two_args)
    )(universal_acceptor)
    assert f.__name__ == "hello_world"


def test_copying_sets_docstring():
    f = define_function_signature(
        "foo", "A docstring for foo", signature(has_two_args)
    )(universal_acceptor)
    assert f.__doc__ == "A docstring for foo"


def test_uses_defaults():
    f = define_function_signature(
        "foo", "A docstring for foo", signature(has_a_default)
    )(universal_acceptor)
    assert f(3, 2) == ((3, 2, 1), {})


def test_uses_varargs():
    f = define_function_signature("foo", "A docstring for foo", signature(has_varargs))(
        universal_acceptor
    )
    assert f(1, 2) == ((1, 2), {})


DEFINE_FOO_FUNCTION = """
def foo(x):
    return x
"""


def test_exec_as_module_execs():
    m = source_exec_as_module(DEFINE_FOO_FUNCTION)
    assert m.foo(1) == 1


def test_exec_as_module_caches():
    assert source_exec_as_module(DEFINE_FOO_FUNCTION) is source_exec_as_module(
        DEFINE_FOO_FUNCTION
    )


def test_exec_leaves_sys_path_unchanged():
    old_path = deepcopy(sys.path)
    source_exec_as_module("hello_world = 42")
    assert sys.path == old_path


def test_define_function_signature_works_with_conflicts():
    def accepts_everything(*args, **kwargs):
        pass

    define_function_signature(
        "hello",
        "A docstring for hello",
        Signature(parameters=[Parameter("f", Parameter.POSITIONAL_OR_KEYWORD)]),
    )(accepts_everything)(1)

    define_function_signature(
        "hello",
        "A docstring for hello",
        Signature(parameters=[Parameter("f", Parameter.VAR_POSITIONAL)]),
    )(accepts_everything)(1)

    define_function_signature(
        "hello",
        "A docstring for hello",
        Signature(parameters=[Parameter("f", Parameter.VAR_KEYWORD)]),
    )(accepts_everything)()

    define_function_signature(
        "hello",
        "A docstring for hello",
        Signature(
            parameters=[
                Parameter("f", Parameter.POSITIONAL_OR_KEYWORD),
                Parameter("f_3", Parameter.POSITIONAL_OR_KEYWORD),
                Parameter("f_1", Parameter.VAR_POSITIONAL),
                Parameter("f_2", Parameter.VAR_KEYWORD),
            ]
        ),
    )(accepts_everything)(1, 2)


def test_define_function_signature_validates_function_name():
    define_function_signature("hello_world", None, Signature())
    with raises(ValueError):
        define_function_signature("hello world", None, Signature())


class Container:
    def funcy(self):
        pass


def test_can_proxy_functions_with_mixed_args_and_varargs():
    def foo(a, *args):
        return (a, args)

    @proxies(foo)
    def bar(*args, **kwargs):
        return foo(*args, **kwargs)

    assert bar(1, 2) == (1, (2,))


def test_can_delegate_to_a_function_with_no_positional_args():
    def foo(a, b):
        return (a, b)

    @proxies(foo)
    def bar(**kwargs):
        return foo(**kwargs)

    assert bar(2, 1) == (2, 1)


@pytest.mark.parametrize(
    "func,args,expected",
    [
        (lambda: None, (), None),
        (lambda a: a**2, (2,), 4),
        (lambda *a: a, [1, 2, 3], (1, 2, 3)),
    ],
)
def test_can_proxy_lambdas(func, args, expected):
    @proxies(func)
    def wrapped(*args, **kwargs):
        return func(*args, **kwargs)

    assert wrapped.__name__ == "<lambda>"
    assert wrapped(*args) == expected


class Snowman:
    def __repr__(self):
        return "☃"


class BittySnowman:
    def __repr__(self):
        return "☃"


def test_can_handle_unicode_repr():
    def foo(x):
        pass

    assert repr_call(foo, [Snowman()], {}) == "foo(x=☃)"
    assert repr_call(foo, [], {"x": Snowman()}) == "foo(x=☃)"


class NoRepr:
    pass


def test_can_handle_repr_on_type():
    def foo(x):
        pass

    assert repr_call(foo, [Snowman], {}) == "foo(x=Snowman)"
    assert repr_call(foo, [NoRepr], {}) == "foo(x=NoRepr)"


def test_can_handle_repr_of_none():
    def foo(x):
        pass

    assert repr_call(foo, [None], {}) == "foo(x=None)"
    assert repr_call(foo, [], {"x": None}) == "foo(x=None)"


def test_kwargs_appear_in_arg_string():
    def varargs(*args, **kwargs):
        pass

    assert "x=1" in repr_call(varargs, (), {"x": 1})


def test_is_mock_with_negative_cases():
    assert not is_mock(None)
    assert not is_mock(1234)
    assert not is_mock(is_mock)
    assert not is_mock(BittySnowman())
    assert not is_mock("foobar")
    assert not is_mock(Mock(spec=BittySnowman))
    assert not is_mock(MagicMock(spec=BittySnowman))


def test_is_mock_with_positive_cases():
    assert is_mock(Mock())
    assert is_mock(MagicMock())
    assert is_mock(NonCallableMock())
    assert is_mock(NonCallableMagicMock())


class Target:
    def __init__(self, a, b):
        pass

    def method(self, a, b):
        pass


@pytest.mark.parametrize("target", [Target, Target(1, 2).method])
@pytest.mark.parametrize(
    "args,kwargs,expected",
    [
        ((), {}, set("ab")),
        ((1,), {}, set("b")),
        ((1, 2), {}, set()),
        ((), {"a": 1}, set("b")),
        ((), {"b": 2}, set("a")),
        ((), {"a": 1, "b": 2}, set()),
    ],
)
def test_required_args(target, args, kwargs, expected):
    # Mostly checking that `self` (and only self) is correctly excluded
    assert required_args(target, args, kwargs) == expected


def test_can_handle_unicode_identifier_in_same_line_as_lambda_def():
    # fmt: off
    pi = "π"; is_str_pi = lambda x: x == pi  # noqa: E702
    # fmt: on
    assert get_pretty_function_description(is_str_pi) == "lambda x: x == pi"


def test_too_many_posargs_fails():
    with pytest.raises(TypeError):
        st.times(time.min, time.max, st.none(), st.none()).validate()


def test_overlapping_posarg_kwarg_fails():
    with pytest.raises(TypeError):
        st.times(time.min, time.max, st.none(), timezones=st.just(None)).validate()


def test_inline_given_handles_self():
    # Regression test for https://github.com/HypothesisWorks/hypothesis/issues/961
    class Cls:
        def method(self, **kwargs):
            assert isinstance(self, Cls)
            assert kwargs["k"] is sentinel

    sentinel = object()
    given(k=st.just(sentinel))(Cls().method)()


def logged(f):
    @wraps(f)
    def wrapper(*a, **kw):
        return f(*a, **kw)

    return wrapper


class Bar:
    @logged
    def __init__(self, i: int):
        pass


@given(st.builds(Bar))
def test_issue_2495_regression(_):
    """See https://github.com/HypothesisWorks/hypothesis/issues/2495"""


@pytest.mark.skipif(
    sys.version_info[:2] >= (3, 11),
    reason="handled upstream in https://github.com/python/cpython/pull/92065",
)
def test_error_on_keyword_parameter_name():
    def f(source):
        pass

    f.__signature__ = Signature(
        parameters=[Parameter("from", Parameter.KEYWORD_ONLY)],
        return_annotation=Parameter.empty,
    )

    with pytest.raises(ValueError, match="SyntaxError because `from` is a keyword"):
        get_signature(f)


def test_param_is_called_within_func():
    def f(any_name):
        any_name()

    assert is_first_param_referenced_in_function(f)


def test_param_is_called_within_subfunc():
    def f(any_name):
        def f2():
            any_name()

    assert is_first_param_referenced_in_function(f)


def test_param_is_not_called_within_func():
    def f(any_name):
        pass

    assert not is_first_param_referenced_in_function(f)


def test_param_called_within_defaults_on_error():
    # Create a function object for which we cannot retrieve the source.
    f = compile("lambda: ...", "_.py", "eval")
    assert is_first_param_referenced_in_function(f)


def _prep_source(*pairs):
    return [
        pytest.param(
            dedent(x).strip(), dedent(y).strip().encode(), id=f"case-{i}", marks=marks
        )
        for i, (x, y, *marks) in enumerate(pairs)
    ]


@pytest.mark.parametrize(
    "src, clean",
    _prep_source(
        ("", ""),
        ("def test(): pass", "def test(): pass"),
        ("def invalid syntax", "def invalid syntax"),
        ("def also invalid(", "def also invalid("),
        (
            """
            @example(1)
            @given(st.integers())
            def test(x):
                # line comment
                assert x  # end-of-line comment


                "Had some blank lines above"
            """,
            """
            def test(x):
                assert x
                "Had some blank lines above"
            """,
        ),
        (
            """
            @dec
            async def f():
                pass
            """,
            """
            async def f():
                pass
            """,
        ),
    ),
)
def test_clean_source(src, clean):
    assert reflection._clean_source(src).splitlines() == clean.splitlines()


def test_overlong_repr_warns():
    with pytest.warns(HypothesisWarning, match="overly large"):
        repr(LazyStrategy(st.one_of, [st.none()] * 10000, {}))


def identity(x):
    return x


class Identity:
    def __call__(self, x):
        return x

    def instance_identity(self, x):
        return x

    def instance_self(self):
        return self

    @staticmethod
    def static_identity(x):
        return x

    @classmethod
    def class_identity(cls, x):
        return x


@pytest.mark.parametrize(
    "f",
    [
        lambda x: x,
        identity,
        Identity(),
        Identity().instance_identity,
        Identity.static_identity,
        Identity.class_identity,
    ],
)
def test_is_identity(f):
    assert is_identity_function(f)


@example(Identity().instance_self)
@example(lambda x, y: x)
@example(lambda: None)
@given(st.functions(like=identity, returns=st.from_type(type).flatmap(st.from_type)))
def test_is_not_identity(f):
    obj = object()
    try:
        assume(f(obj) is not obj)  # not necessary but ...
    except TypeError:
        pass
    assert not is_identity_function(f)


@pytest.mark.parametrize(
    "f",
    [
        lambda x, y=None: x,  # this one would be fairly easy to recognize
        lambda x, y=None: y or x,  # ... but the general case is impossible
    ],
)
def test_is_unrecognized_identity(f):
    assert not is_identity_function(f)


def test_cache_key_size_is_bounded():
    # Modify co_consts because ("c" * 1000) may not be evaluated at compile time
    f = lambda x="a" * 1000, *, y="b" * 1000: "c"
    f.__code__ = f.__code__.replace(
        co_consts=tuple(c * 1000 if c == "c" else c for c in f.__code__.co_consts)
    )
    assert len(repr(lambda_sources._function_key(f))) > 3000
    assert len(repr(lambda_sources._function_key(f, bounded_size=True))) < 1000


def test_function_key_distinguishes_alpha_renames():
    # these terms are equivalent under the lambda calculus, but their
    # representations are not, so they should be cached differently.
    assert lambda_sources._function_key(lambda x: x) != lambda_sources._function_key(
        lambda y: y
    )


def test_import():
    import time as t

    f = lambda: t.ctime()
    assert get_pretty_function_description(f) == "lambda: t.ctime()"


@pytest.mark.parametrize("nop_on_f", [True, False])
def test_code_normalization(nop_on_f):
    f = lambda x: x
    g = lambda x: x
    h = f if nop_on_f else g
    assert lambda_sources._function_key(f) == lambda_sources._function_key(g)

    # Append a NOP to one of the bytecodes
    NOP_instr = bytes([lambda_sources._op.NOP, 0])
    h.__code__ = h.__code__.replace(co_code=h.__code__.co_code + NOP_instr)
    assert lambda_sources._function_key(f) != lambda_sources._function_key(g)

    # ...and then normalize g to match f (adding or removing a NOP)
    g.__code__ = lambda_sources._normalize_code(f, g)
    assert lambda_sources._function_key(f) == lambda_sources._function_key(g)


@pytest.mark.parametrize(
    "f, source",
    [
        (lambda x=1, *, y=2: (x, y), "lambda x=1, *, y=2: (x, y)"),
        (lambda x=1, *, y: (x, y), "lambda x=1, *, y: (x, y)"),
        (lambda x, *, y=2: (x, y), "lambda x, *, y=2: (x, y)"),
        (lambda x, *, y: (x, y), "lambda x, *, y: (x, y)"),
        (lambda *, y=2: (x, y), "lambda *, y=2: (x, y)"),  # noqa: F821
        (lambda *, y: (x, y), "lambda *, y: (x, y)"),  # noqa: F821
        (lambda x, /, y=1: (x, y), "lambda x, /, y=1: (x, y)"),
        (lambda x, /, y: (x, y), "lambda x, /, y: (x, y)"),
        (lambda x, /: (x, y), "lambda x, /: (x, y)"),  # noqa: F821
        (lambda x=1, /, y=2: (x, y), "lambda x=1, /, y=2: (x, y)"),
        (lambda x=1, /: (x, y), "lambda x=1, /: (x, y)"),  # noqa: F821
    ],
)
def test_lambda_mimicry_with_arg_defaults(f, source):
    assert get_pretty_function_description(f) == source



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import attrs

from hypothesis.vendor import pretty


class ReprDetector:
    def _repr_pretty_(self, p, cycle):
        """Exercise the IPython callback interface."""
        p.text("GOOD")

    def __repr__(self):
        return "BAD"


@attrs.define
class SomeAttrsClass:
    x: ReprDetector


def test_pretty_prints_attrs_classes():
    assert pretty.pretty(SomeAttrsClass(ReprDetector())) == "SomeAttrsClass(x=GOOD)"


@attrs.define
class SomeAttrsClassWithCustomPretty:
    def _repr_pretty_(self, p, cycle):
        """Exercise the IPython callback interface."""
        p.text("I am a banana")


def test_custom_pretty_print_method_overrides_field_printing():
    assert pretty.pretty(SomeAttrsClassWithCustomPretty()) == "I am a banana"


@attrs.define
class SomeAttrsClassWithLotsOfFields:
    a: int
    b: int
    c: int
    d: int
    e: int
    f: int
    g: int
    h: int
    i: int
    j: int
    k: int
    l: int
    m: int
    n: int
    o: int
    p: int
    q: int
    r: int
    s: int


def test_will_line_break_between_fields():
    obj = SomeAttrsClassWithLotsOfFields(
        **{
            at.name: 12345678900000000000000001
            for at in SomeAttrsClassWithLotsOfFields.__attrs_attrs__
        }
    )
    assert "\n" in pretty.pretty(obj)


@attrs.define
class SomeDataClassWithNoFields: ...


def test_prints_empty_dataclass_correctly():
    assert pretty.pretty(SomeDataClassWithNoFields()) == "SomeDataClassWithNoFields()"


@attrs.define
class AttrsClassWithNoInitField:
    x: int
    y: int = attrs.field(init=False)


def test_does_not_include_no_init_fields_in_attrs_printing():
    record = AttrsClassWithNoInitField(x=1)
    assert pretty.pretty(record) == "AttrsClassWithNoInitField(x=1)"
    record.y = 1
    assert pretty.pretty(record) == "AttrsClassWithNoInitField(x=1)"


class Namespace:
    @attrs.define
    class A:
        x: int


def test_includes_namespace_classes_in_pretty():
    obj = Namespace.A(x=1)
    assert pretty.pretty(obj) == "Namespace.A(x=1)"



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import math

from hypothesis import given, strategies as st
from hypothesis.internal.compat import ceil, floor, int_from_bytes, int_to_bytes


@given(st.binary())
def test_convert_back(bs):
    bs = bytearray(bs)
    assert int_to_bytes(int_from_bytes(bs), len(bs)) == bs


bytes8 = st.builds(bytearray, st.binary(min_size=8, max_size=8))


@given(bytes8, bytes8)
def test_to_int_in_big_endian_order(x, y):
    x, y = sorted((x, y))
    assert 0 <= int_from_bytes(x) <= int_from_bytes(y)


ints8 = st.integers(min_value=0, max_value=2**63 - 1)


@given(ints8, ints8)
def test_to_bytes_in_big_endian_order(x, y):
    x, y = sorted((x, y))
    assert int_to_bytes(x, 8) <= int_to_bytes(y, 8)


@given(st.fractions())
def test_ceil(x):
    assert isinstance(ceil(x), int)
    assert x <= ceil(x) < x + 1
    assert ceil(x) == math.ceil(x)


@given(st.fractions())
def test_floor(x):
    assert isinstance(floor(x), int)
    assert x - 1 < floor(x) <= x
    assert floor(x) == math.floor(x)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from hypothesis import given, strategies as st


class BadRepr:
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return self.value


Frosty = BadRepr("☃")


def test_just_frosty():
    assert repr(st.just(Frosty)) == "just(☃)"


def test_sampling_snowmen():
    assert repr(st.sampled_from((Frosty, "hi"))) == "sampled_from((☃, 'hi'))"


def varargs(*args, **kwargs):
    pass


@given(
    st.sampled_from(
        [
            "✐",
            "✑",
            "✒",
            "✓",
            "✔",
            "✕",
            "✖",
            "✗",
            "✘",
            "✙",
            "✚",
            "✛",
            "✜",
            "✝",
            "✞",
            "✟",
            "✠",
            "✡",
            "✢",
            "✣",
        ]
    )
)
def test_sampled_from_bad_repr(c):
    pass



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import unittest

import pytest
from _pytest.outcomes import Skipped

from hypothesis import given, settings
from hypothesis.core import skip_exceptions_to_reraise
from hypothesis.database import InMemoryExampleDatabase
from hypothesis.strategies import integers

from tests.common.utils import capture_out


@pytest.mark.parametrize("skip_exception", skip_exceptions_to_reraise())
def test_no_falsifying_example_if_unittest_skip(skip_exception):
    """If a ``SkipTest`` exception is raised during a test, Hypothesis should
    not continue running the test and shrink process, nor should it print
    anything about falsifying examples."""

    class DemoTest(unittest.TestCase):
        @given(xs=integers())
        def test_to_be_skipped(self, xs):
            if xs == 0:
                raise skip_exception
            else:
                assert xs == 0

    with capture_out() as o:
        suite = unittest.defaultTestLoader.loadTestsFromTestCase(DemoTest)
        unittest.TextTestRunner().run(suite)

    assert "Falsifying example" not in o.getvalue()


def test_skip_exceptions_save_database_entries():
    """Skip exceptions should save database entries for immediate replay (issue #4484)."""
    database = InMemoryExampleDatabase()
    call_count = 0
    skip_value = None
    first_value = None

    @settings(database=database, max_examples=100)
    @given(integers())
    def test_func(n):
        nonlocal call_count, skip_value, first_value
        call_count += 1
        if first_value is None:
            first_value = n

        # Skip on the 5th value in the first run (the choice of 5 is arbitrary)
        if call_count == 5 and skip_value is None:
            skip_value = n

        if n == skip_value:
            pytest.skip()

    # First run should raise a skip exception and save to database
    with pytest.raises(Skipped):
        test_func()
    assert sum(len(v) for v in database.data.values()) == 1

    # Second run should immediately replay the skip value
    first_value = None
    call_count = 0
    with pytest.raises(Skipped):
        test_func()

    # first call should be the replayed skip value
    assert first_value == skip_value
    assert call_count == 1



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import pytest

from hypothesis import given, strategies as st


@pytest.mark.parametrize("width", [32, 64, 128])
@pytest.mark.parametrize("keyword", ["min_magnitude", "max_magnitude"])
@given(data=st.data())
def test_magnitude_validates(width, keyword, data):
    # See https://github.com/HypothesisWorks/hypothesis/issues/3573
    component_width = width / 2
    magnitude = data.draw(
        # 1.8 is a known example that hasn't validated in the past
        st.floats(0, width=component_width) | st.just(1.8),
        label=keyword,
    )
    strat = st.complex_numbers(width=width, **{keyword: magnitude})
    data.draw(strat)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import unittest

import pytest
from _pytest.outcomes import Failed, Skipped

from hypothesis import Phase, example, find, given, reject, settings, strategies as st
from hypothesis.database import InMemoryExampleDatabase
from hypothesis.errors import InvalidArgument, NoSuchExample, Unsatisfiable


def test_stops_after_max_examples_if_satisfying():
    count = 0

    def track(x):
        nonlocal count
        count += 1
        return False

    max_examples = 100

    with pytest.raises(NoSuchExample):
        find(st.integers(0, 10000), track, settings=settings(max_examples=max_examples))

    assert count == max_examples


def test_stops_after_ten_times_max_examples_if_not_satisfying():
    count = 0

    def track(x):
        nonlocal count
        count += 1
        reject()

    max_examples = 100
    with pytest.raises(Unsatisfiable):
        find(st.integers(0, 10000), track, settings=settings(max_examples=max_examples))

    # Very occasionally we can generate overflows in generation, which also
    # count towards our example budget, which means that we don't hit the
    # maximum.
    assert count <= 10 * max_examples


some_normal_settings = settings()


def test_is_not_normally_default():
    assert settings.default is not some_normal_settings


@given(st.booleans())
@some_normal_settings
def test_settings_are_default_in_given(x):
    assert settings.default is some_normal_settings


def test_given_shrinks_pytest_helper_errors():
    final_value = None

    @settings(derandomize=True, max_examples=100)
    @given(st.integers())
    def inner(x):
        nonlocal final_value
        final_value = x
        if x > 100:
            pytest.fail(f"{x=} is too big!")

    with pytest.raises(Failed):
        inner()
    assert final_value == 101


def test_pytest_skip_skips_shrinking():
    seen_large = False

    @settings(derandomize=True, max_examples=100)
    @given(st.integers())
    def inner(x):
        nonlocal seen_large
        if x > 100:
            if seen_large:
                raise Exception("Should never replay a skipped test!")
            seen_large = True
            pytest.skip(f"{x=} is too big!")

    with pytest.raises(Skipped):
        inner()


def test_can_find_with_db_eq_none():
    find(st.integers(), bool, settings=settings(database=None, max_examples=100))


def test_no_such_example():
    with pytest.raises(NoSuchExample):
        find(st.none(), bool, database_key=b"no such example")


def test_validates_strategies_for_test_method():
    invalid_strategy = st.lists(st.nothing(), min_size=1)

    class TestStrategyValidation:
        @given(invalid_strategy)
        def test_method_with_bad_strategy(self, x):
            pass

    instance = TestStrategyValidation()
    with pytest.raises(InvalidArgument):
        instance.test_method_with_bad_strategy()


@example(1)
@given(st.integers())
@settings(phases=[Phase.target, Phase.shrink, Phase.explain])
def no_phases(_):
    raise Exception


@given(st.integers())
@settings(phases=[Phase.explicit])
def no_explicit(_):
    raise Exception


@given(st.integers())
@settings(phases=[Phase.reuse], database=InMemoryExampleDatabase())
def empty_db(_):
    raise Exception


@pytest.mark.parametrize(
    "test_fn",
    [no_phases, no_explicit, empty_db],
    ids=lambda t: t.__name__,
)
def test_non_executed_tests_raise_skipped(test_fn):
    with pytest.raises(unittest.SkipTest):
        test_fn()


@pytest.mark.parametrize(
    "codec, max_codepoint, exclude_categories, categories",
    [
        ("ascii", None, None, None),
        ("ascii", 128, None, None),
        ("ascii", 100, None, None),
        ("utf-8", None, None, None),
        ("utf-8", None, ["Cs"], None),
        ("utf-8", None, ["N"], None),
        ("utf-8", None, None, ["N"]),
    ],
)
@given(st.data())
def test_characters_codec(codec, max_codepoint, exclude_categories, categories, data):
    strategy = st.characters(
        codec=codec,
        max_codepoint=max_codepoint,
        exclude_categories=exclude_categories,
        categories=categories,
    )
    example = data.draw(strategy)
    assert example.encode(encoding=codec).decode(encoding=codec) == example



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import math

import pytest

import hypothesis.strategies as st
from hypothesis import HealthCheck, example, given, settings
from hypothesis.internal.conjecture import utils as cu
from hypothesis.internal.conjecture.choice import choice_equal
from hypothesis.internal.conjecture.data import ConjectureData
from hypothesis.internal.floats import SIGNALING_NAN, SMALLEST_SUBNORMAL

from tests.conjecture.common import choice_types_constraints, fresh_data


@given(st.data())
@settings(
    database=None,
    suppress_health_check=[HealthCheck.filter_too_much, HealthCheck.too_slow],
)
def test_forced_many(data):
    forced = data.draw(st.integers(0, 100))
    min_size = data.draw(st.integers(0, forced))
    max_size = data.draw(st.integers(forced, 100))
    assert min_size <= forced <= max_size  # by construction

    data = fresh_data()
    many = cu.many(
        data,
        min_size=min_size,
        average_size=(min_size + max_size) / 2,
        max_size=max_size,
        forced=forced,
    )
    for _ in range(forced):
        assert many.more()

    assert not many.more()

    # ensure values written to the buffer do in fact generate the forced value
    data = ConjectureData.for_choices(data.choices)
    many = cu.many(
        data,
        min_size=min_size,
        average_size=(min_size + max_size) / 2,
        max_size=max_size,
    )
    for _ in range(forced):
        assert many.more()

    assert not many.more()


@example(("boolean", {"p": 1e-19, "forced": True}))  # 64 bit p
@example(("boolean", {"p": 3e-19, "forced": True}))  # 62 bit p
@example(
    (
        "integer",
        {
            "min_value": -1,
            "max_value": 1,
            "shrink_towards": 1,
            "weights": {-1: 0.2, 0: 0.2, 1: 0.2},
            "forced": 0,
        },
    )
)
@example(
    (
        "integer",
        {
            "min_value": -1,
            "max_value": 1,
            "shrink_towards": -1,
            "weights": {-1: 0.2, 0: 0.2, 1: 0.2},
            "forced": 0,
        },
    )
)
@example(
    (
        "integer",
        {
            "min_value": 10,
            "max_value": 1_000,
            "shrink_towards": 17,
            "weights": {20: 0.1},
            "forced": 15,
        },
    )
)
@example(
    (
        "integer",
        {
            "min_value": -1_000,
            "max_value": -10,
            "shrink_towards": -17,
            "weights": {-20: 0.1},
            "forced": -15,
        },
    )
)
@example(("float", {"forced": 0.0}))
@example(("float", {"forced": -0.0}))
@example(("float", {"forced": 1.0}))
@example(("float", {"forced": 1.2345}))
@example(("float", {"forced": SMALLEST_SUBNORMAL}))
@example(("float", {"forced": -SMALLEST_SUBNORMAL}))
@example(("float", {"forced": 100 * SMALLEST_SUBNORMAL}))
@example(("float", {"forced": math.nan}))
@example(("float", {"forced": -math.nan}))
@example(("float", {"forced": SIGNALING_NAN}))
@example(("float", {"forced": -SIGNALING_NAN}))
@example(("float", {"forced": 1e999}))
@example(("float", {"forced": -1e999}))
# previously errored on our {pos, neg}_clamper logic not considering nans.
@example(
    (
        "float",
        {"min_value": -1 * math.inf, "max_value": -1 * math.inf, "forced": math.nan},
    )
)
@given(choice_types_constraints(use_forced=True))
def test_forced_values(choice_type_and_constraints):
    (choice_type, constraints) = choice_type_and_constraints
    constraints = constraints.copy()
    forced = constraints["forced"]
    data = fresh_data()
    assert choice_equal(getattr(data, f"draw_{choice_type}")(**constraints), forced)

    # now make sure the written buffer reproduces the forced value, even without
    # specifying forced=.
    del constraints["forced"]
    data = ConjectureData.for_choices(data.choices)
    assert choice_equal(getattr(data, f"draw_{choice_type}")(**constraints), forced)


@pytest.mark.parametrize("sign", [1, -1])
@pytest.mark.parametrize(
    "min_value, max_value",
    [
        (0.0, 0.0),
        (-0.0, -0.0),
        (0.0, 100.0),
        (-100.0, -0.0),
        (5.0, 10.0),
        (-10.0, -5.0),
    ],
)
@given(random=st.randoms())
def test_forced_floats_with_nan(random, sign, min_value, max_value):
    # nans with a sign opposite of both bounds previously gave us trouble
    # trying to use float clampers that didn't exist when drawing.
    data = fresh_data(random=random)
    data.draw_float(min_value=min_value, max_value=max_value, forced=sign * math.nan)


@given(st.data())
def test_forced_with_large_magnitude_integers(data):
    bound_offset = data.draw(st.integers(min_value=0))
    # forced_offset = bound_offset + st.integers(min_value=0) may look cleaner, but
    # has subtly different maximum value semantics as it is twice the range of a
    # single draw
    forced_offset = data.draw(st.integers(min_value=bound_offset))

    half_range = 2**127 + 1
    cd = fresh_data()
    cd.draw_integer(
        min_value=half_range + bound_offset, forced=half_range + forced_offset
    )

    cd = fresh_data()
    cd.draw_integer(
        max_value=-(half_range + bound_offset), forced=-(half_range + forced_offset)
    )
